from . import target_base
from . import utils
from . import dirs
from . import db
from . import col
import os.path
join = os.path.join

pypkgname = __name__.split('.')[0]
assert pypkgname != '__main__'

def _ensure_create_pyinit(pkg,compinit):
    d  = dirs.pkg_cache_dir(pkg,'pyinit')
    fn = join(d,('__init__ci.py' if compinit else '__init__.py'))
    if not os.path.exists(fn):
        cont = '#autogenerated file\n"""python module for package %s"""\n'%pkg.name
        if compinit:
            cont+='from . _init import *\n'
        utils.mkdir_p(d)
        open(fn,'w').write(cont)
    return fn

class TargetPyInit(target_base.Target):
    def __init__(self,pkg):
        compinit = os.path.isdir(dirs.pkg_dir(pkg,'pycpp__init'))
        fn=_ensure_create_pyinit(pkg,compinit)
        db.db['pkg2parts'][pkg.name].add('autopyinit__%s'%os.path.basename(fn))
        neededlinks=set([(dirs.installdir / 'python' / pkg.name,'__init__.py',fn)])
        pklfile=dirs.pkg_cache_dir(pkg,'pyinit','pyinit.pkl')
        mpf=dirs.makefile_blddir(pklfile)
        utils.update_pkl_if_changed(neededlinks,pklfile)

        self.name='%s_pyinit'%pkg.name
        self.pkgname=pkg.name
        self.deps=[mpf]
        self.code = ['@if [ ${VERBOSE} -ge 0 ]; then echo "%s  Generating %s/__init__.py%s"; fi'%(col.bldcol('pymod'),pkg.name,col.bldend),
                     'python3 -m%s.instsl2 %s ${VERBOSE}'%(pypkgname,mpf)]

_python_patterns = set(['pycp', 'pyth'])#todo: to conf
def tfactory_pyinit(pkg,dirtypes):
    #if we have pure or compiled python modules and we do not have a
    #user written __init__.py, we will have to add one. That one will
    #either be empty, or include the compiled pycpp__init module.
    if dirtypes.isdisjoint(_python_patterns):
        return []
    if 'pyth' in dirtypes and os.path.exists(dirs.pkg_dir(pkg,'python','__init__.py')):
        return []#user supplied
    return [TargetPyInit(pkg)]




#cleanup: only needed when a package itself is touched, i.e. can take place in target init or in target factories.
#
#issues: remove symlinked scripts, python, data, ...
#        remove library which is no longer created
#        remove app which is no longer created
#
#        can we add a cleanup target? Which is only triggered once after pkg targets are recreated.



from . import utils

def cluster_ncpus():
    #if running at cluster via slurm, respect SLURM_CPUS_PER_TASK limit
    from os import environ as env
    cname = env.get('SLURM_CLUSTER_NAME','').strip()
    cncpu = env.get('SLURM_CPUS_PER_TASK','notfound').strip()
    if cname.lower()=='dmsc' and cncpu.isdigit():
        return int(cncpu)

def get_n_cores():
    import os
    if os.path.exists('/proc/cpuinfo'):
        #linux
        n=0
        for ll in open('/proc/cpuinfo'):
            if ( ll.startswith('processor')
                 and ll.split()[0:2]==['processor',':'] ):
                n += 1
        if not n:
            from .error import warn
            warn('Warning: Could not determine number of processors.'
                 ' Assuming just one present (override with -jN)')
            return 1
        return n
    else:
        #osx
        (ec,n)=utils.run('sysctl -n hw.ncpu')
        if ec:
            from .error import warn
            warn('Warning: Could not determine number of processors.'
                 ' Assuming just one present (override with -jN)')
            return 1
        return int(n.strip())

def get_load(ncores):
    (ec,p)=utils.run('/bin/ps -eo pcpu')
    if ec:
        from .error import warn
        warn('Warning: Could not determine current CPU load. Assuming 0%.')
        return 0.0
    if hasattr(p,'decode'):
        p=p.decode('ascii')
    p=0.01*sum([float(x.replace(',','.')) for x in p.split()[1:]])
    if p>ncores*1.5:
        from .error import warn
        warn('Warning: Could not determine current CPU load'
             ' ("ps -eo pcpu" output was suspicous). Assuming 0%.')
        return 0.0
    from os import environ as env
    if env.get('GITHUB_SERVER_URL',''): #Ignore CPU load for better performance
                                        #when using GitHub Runners (CI)
      return 0.0
    return p

def auto_njobs():
    n_reserved = cluster_ncpus()
    if n_reserved:
        return n_reserved
    nc=get_n_cores()
    freecores=nc-get_load(nc)
    return max(1,int(0.5+0.98*freecores))




def load_env_cache( *keys ):
    import simplebuild.cfg as cfg
    import pickle
    with cfg.dirs.envcache.open('rb') as fh:
        data = pickle.load(fh)
    for k in keys:
        data = data[k]
    return data

def find_libpython():
    import shlex
    candidates=[]
    linkflags = load_env_cache('system',
                               'general',
                               'pybind11_embed_linkflags_list')
    allflags = []
    for lf in shlex.split(linkflags):
        for e in lf.split(';'):
            if e.strip():
                allflags.append(e.strip())
    candidates = [e for e in allflags
                  if '.so' in e]
    if len(candidates)>1:
        candidates = [e for e in candidates
                      if 'python' in e.lower()]
    if len(candidates)>1:
        candidates = [e for e in candidates
                      if 'libpython' in e.lower()]
    if len(candidates)==1:
        import pathlib
        p = pathlib.Path(candidates[0])
        if p.is_file():
            return p
    raise RuntimeError('Could not extract libpython location'
                       f' from pybind11 linkflags: {linkflags}')

if __name__ == '__main__':
    try:
        print( find_libpython() )
    except RuntimeError as e:
        raise SystemExit(str(e))



def decode_pattern(p):
    #pattern like '<delim><pat1><delim><pat2><delim>' (example "/bla/lala/" or "#bla#lala#"
    if len(p)<4 or p.count(p[0])!=3 or p[-1]!=p[0]:
        return None,None#error
    else:
        p=p.split(p[0])
        if len(p)!=4 or p[0] or p[-1]:
            return None,None#error
    return p[1],p[2]#search_pattern,replace_pattern

def replacefile(filename,printname,search_pattern,replace_pattern):
    n=0
    fh=open(filename,'rt')
    cont=''
    try:
        for ll in fh:
            if search_pattern in ll:
                n += ll.count(search_pattern)
                ll = ll.replace(search_pattern,replace_pattern)
            cont += ll
    except UnicodeDecodeError:
        #ignore non-text data files
        n=0
    fh.close()
    if n:
        #at least one replacement made
        fh=open(filename,'wt')
        fh.write(cont)
        fh.close()
        from .io import print
        print('  %50s : %i replacements'%(printname,n))
    return n

def replace(pkg,search_pattern,replace_pattern,filenames):
    from .grep import pkgfiles
    import os
    import pathlib
    pjoin=os.path.join
    n=0
    for f in pkgfiles(pkg):
        if filenames and pathlib.Path(f) not in filenames:
            continue
        n+=replacefile(pjoin(pkg.dirname,f),pjoin(pkg.name,f),search_pattern,replace_pattern)
    return n



from . import target_base
from . import includes
from . import utils
from . import dirs
from . import db
import os
from . import langs

class TargetIncDep(target_base.Target):
    #for header files not part of an inderdependent circular group or for such a circular group.
    def __init__(self,pkg,subdir,fn,pub,priv,possible_privincs,igroup=None):
        self.pkglevel = False#does not need a pkglevel dep, just needs to be present as possible dependency
        isgroup = (igroup is not None)
        assert bool(isgroup)==bool(isinstance(fn,set))
        self.name='%s__%s__%s'%(pkg.name,subdir,('group%i'%igroup if isgroup else fn))
        self.pkgname=pkg.name
        self.code = []

        # NOTE THAT FOLLOWING LINES OF CODE ARE SIMILAR TO ONES IN tfact_binary.py !!!

        #deal with the private dependencies as well as public deps to same pkg
        #here, postpone the other public ones until setup() when other packages are available.
        self.__pklfile=dirs.pkg_cache_dir(pkg,('%s__group%i'%(subdir,igroup) if isgroup else '%s__%s'%(subdir,fn)))
        self.deps=[dirs.makefile_blddir(self.__pklfile)]
        for f in (fn if isgroup else [fn]):
            self.deps += [dirs.makefile_pkg_dir(pkg,subdir,f)]#only actual file dependencies
        self._pkloffset=len(self.deps)
        if priv and possible_privincs:
            for p in priv:
                if p in possible_privincs:
                    self.deps+=['%s__%s__%s'%(pkg.name,subdir,p)]
        if pub:
            pkg.update_incs()
            incs=db.db['pkg2inc'][pkg.name]
            self.deps += ['%s__libinc__%s'%(p,f) for (p,f) in pub if (p==self.pkgname and f in incs)]
            self._pub = [(p,f) for (p,f) in pub if p!=self.pkgname]
        else:
            self._pub=None

    def setup(self,pkg):
        # NOTE THAT FOLLOWING LINES OF CODE ARE SIMILAR TO ONES IN tfact_binary.py !!!
        if self._pub:
            p2i=db.db['pkg2inc']
            d=pkg.deps_names()
            self.deps+=['%s__libinc__%s'%(p,f) for (p,f) in self._pub if (p in d and f in p2i.get(p,[]))]
        pklcont=set(self.deps[self._pkloffset:])
        utils.mkdir_p(os.path.dirname(self.__pklfile))
        utils.update_pkl_if_changed(pklcont,self.__pklfile)

class TargetIncDep_HeaderFileInGroup(target_base.Target):
    #For a single header file in a circular group
    def __init__(self,pkg,subdir,fn,igroup):
        self.pkglevel = False#does not need a pkglevel dep, just needs to be present as possible dependency
        self.name='%s__%s__%s'%(pkg.name,subdir,fn)
        self.pkgname=pkg.name
        self.code = []
        self.deps = ['%s__%s__group%i'%(pkg.name,subdir,igroup)]

def find_circ(hh,hh2local,ll,circgroups):
    local = hh2local[hh]
    if not local:
        return#no unresolved local includes
    for hhl in local:
        if hhl in ll:
            #New circular group! Some circular groups have multiple circular routes within
            #the group, so we must check for overlap with existing groups and merge:
            newcg=set(ll[ll.index(hhl):])
            overlaps=[]
            for cg in circgroups:
                if not cg.isdisjoint(newcg):
                    overlaps+=[cg]
                    newcg.update(cg)
            for ol in overlaps:
                circgroups.remove(ol)
            if len(newcg)>1:#don't treat files with self-includes as groups (let the compiler complain later)
                circgroups+=[newcg]
        else:
            #continue with next layer of includes:
            find_circ(hhl,hh2local,ll+[hhl],circgroups)
    hh2local[hh]=None#prevent recheck

def tfactory_headerdeps(pkg,subdir):
    hhs = langs.headers_in_dir(dirs.pkg_dir(pkg,subdir))
    hh2deps=dict((hh,includes.find_includes(dirs.pkg_dir(pkg,subdir,hh),pkg)) for hh in hhs)
    if not hh2deps and subdir=='libinc':
        from . import error
        error.error("No header files found in %s/%s (remove this directory if not needed)"%(pkg.name,subdir))
    #It is possible to have valid circular dependencies between header files in
    #the same dir. To solve this we must first detect such groups of interdependant files and
    #make special circular targets in the makefile:
    if subdir!='libinc':
        hh2local=dict((hh,hh2deps[hh][0]) for hh in hhs)
    else:
        hh2local={}
        for hh in hhs:
            deps= hh2deps.get(hh,(None,set()))[1]
            hh2local[hh]=set(fn for (pkgname,fn) in deps if pkgname==pkg.name) if deps else set()
    for hh,local in hh2local.items():
        if local:
            local.intersection_update(hhs)
    circgroups=[]
    for hh in hhs:
        find_circ(hh,hh2local,[],circgroups)
    hh2group={}
    for ig,cg in enumerate(circgroups):
        for hh in cg:
            hh2group[hh]=ig

    igroup2pub = {}
    igroup2priv = {}
    tgts=[]
    for hh,deps in hh2deps.items():
        igroup = hh2group.get(hh,None)
        if igroup is None:
            possible_privincs = hhs.difference(set([hh]))
            priv,pub=deps
            if subdir=='libinc':
                if pub:
                    pub = [(p,f) for (p,f) in pub if f!=hh]
            else:
                if priv and hh in priv:
                    priv.remove(hh)
            tgts+=[TargetIncDep(pkg,subdir,hh,pub,priv,possible_privincs)]
        else:
            tgts+=[TargetIncDep_HeaderFileInGroup(pkg,subdir,hh,igroup)]
            if deps[0]:
                if igroup not in igroup2priv:
                    igroup2priv[igroup]=deps[0]
                else:
                    igroup2priv[igroup].update(deps[0])
            if deps[1]:
                if igroup not in igroup2pub:
                    igroup2pub[igroup]=deps[1]
                else:
                    igroup2pub[igroup].update(deps[1])
    for igroup,group in enumerate(circgroups):
        priv=igroup2priv.get(igroup,None)
        pub=igroup2pub.get(igroup,None)
        possible_privincs = hhs.difference(set(group))
        if pub and subdir=='libinc':
            pub = [(p,f) for (p,f) in pub if f not in group]
        tgts+=[TargetIncDep(pkg,subdir,group,pub,priv,possible_privincs,igroup)]
    return tgts



from . import conf

def dotgen(pkgloader,outfile,enabled_only=True):
    fh=open(outfile,'w')
    fh.write("digraph GG {\n")
    fh.write("node [\n")
    fh.write(" fontsize = \"12\"\n")
    fh.write("];\n")
    autodeps=conf.autodeps
    def sortfunc(p):
        n=len(p.direct_clients)+len(p.direct_deps_pkgnames)
        if p.name in autodeps:
            n=-1
        return (n,p.name)
    for p in sorted(pkgloader.pkgs,key=sortfunc):
        if p.name in autodeps:
            assert p.enabled
            fh.write('    "node_%s" [  label="%s" shape="doubleoctagon" ];\n'%(p.name,p.name))
        elif p.enabled:
            fh.write('    "node_%s" [  label="%s" shape="octagon" ];\n'%(p.name,p.name))
        else:
            if not enabled_only:
                fh.write('    "node_%s" [ fontcolor="gray" color="grey" label="%s" shape="octagon" ];\n'%(p.name,p.name))

    for p in pkgloader.pkgs:
        if enabled_only and not p.enabled:
            continue
        for pdname in p.direct_deps_pkgnames:
            if pdname not in autodeps:
                fh.write('    "node_%s" -> "node_%s"%s\n'%(p.name,pdname,'' if p.enabled else ' [ color="gray" ]'))
    fh.write("}\n")
    fh.close()



#!/usr/bin/env python3
def perform_tests(testdir,
                  installdir, njobs, nexcerpts, filters,
                  do_pycoverage, pkgloader=None):
    import os
    import sys
    import fnmatch
    import shutil
    import shlex
    import pathlib
    from .conf import runnable_is_test
    from .utils import mkdir_p
    from .utils import system
    from . import testxml
    from . import dirs
    from . import io as _io
    _ospathjoin=os.path.join

    print = _io.print

    from . import col as _col
    col_ok = _col.ok
    col_bad = _col.bad
    col_end = _col.end

    print(f'Running tests in {testdir}:')
    print()
    sys.stdout.flush()
    sys.stderr.flush()

    assert not os.path.exists(testdir)
    testdir=os.path.abspath(testdir)
    line_hr = " ---------------------------------------+-----------+--------+----------+------------------"
    line_titles = "  Test job results                      | Time [ms] | Job EC | Log-diff | Trouble info"
    def print_header():
        print(line_hr)
        print(line_titles)
        print(line_hr)
    def print_footer():
        print(line_hr)

    logdir=_ospathjoin(installdir,'tests/testref')
    #bindirs=[_ospathjoin(installdir,'bin'),_ospathjoin(installdir,'scripts')]
    logfiles = set(fn for fn in os.listdir(logdir) if fn.endswith('.log')) if os.path.exists(logdir) else set()

    def ldtest(d):
        if os.path.exists(d):
            for f in os.listdir(d):
                if os.path.isdir(_ospathjoin(d,f)):
                    continue#skip some weird .DSYM dirs for fortran apps on osx
                if f+'.log' in logfiles or runnable_is_test(f):
                    yield _ospathjoin(d,f)
    tests=set()
    testinfo = {}
    #Load pkg list from the generated simplebuild.cfg module, for complete
    #consistency between "simplebuild -t" and "sbtests" commands (TODO obsolete
    #requirement?):
    if pkgloader:
        #Called from frontend.py
        pkginfo = dict((p.name,p.info_as_dict()) for p in pkgloader.pkgs)
    else:
        #called from CLI sbtests:
        #FIXME: supported?
        sys.path.insert(0,str(dirs.installdir/'python'))
        import simplebuild.cfg as simplebuild_cfg#FIXME: Could/should we brute-force load via importlib, so avoid poluting the module space here?
        sys.path.pop(0)
        pkginfo = simplebuild_cfg.pkgs

    if filters:
        # Use same logic as for the pkg_filters, i.e.:

        # * If any positive filters are defined, a pkg must pass at least one of
        #   them to pass.
        # * If any negative filters are defined, a pkg must NOT pass any of
        #   them, to pass.

        _pos_filters = []
        _neg_filters = []
        for fltr in filters:
            fltr=fltr.strip()
            if fltr.startswith('!'):
                fltr = fltr[1:].strip()
                if fltr:
                    _neg_filters.append( fltr )
            elif fltr:
                _pos_filters.append( fltr )
        def filter_fct( testname ):
            if _pos_filters:
                if not any( fnmatch.fnmatch(runnable,fltr)
                            for fltr in _pos_filters ):
                    return False
            if _neg_filters:
                if any( fnmatch.fnmatch(runnable,fltr)
                        for fltr in _neg_filters ):
                    return False
            return True
    else:
        def filter_fct( testname ):
            return True


    nfiltered = 0
    for p in (p for _,p in pkginfo.items() if p['enabled']):
        reflogs, runnables = p['reflogs'],p['runnables']
        for runnable in runnables:
            hasreflog = '%s.log'%runnable in reflogs
            if hasreflog or runnable_is_test(runnable):
                if not filter_fct( runnable ):
                    nfiltered += 1
                else:
                    assert runnable not in tests
                    tests.add(runnable)
                    testinfo[runnable] = {'pkgname':p['name']}

    if not tests:
        if nfiltered:
            print(f'WARNING: All {nfiltered} tests were blocked by specified filters.')
            print()
        print_header()
        print('  n/a')
        print_footer()
        return 0

    mkdir_p(testdir)
    mf=open(_ospathjoin(testdir,'Makefile'),'w')

    mf.write('TESTDIR:=%s\n\n'%os.path.abspath(testdir))

    mf.write('.DEFAULT: all\n\n')
    mf.write('.PHONY: all\n\n')

    alltests=[]

    if do_pycoverage:
        coverage_out_dir = pathlib.Path(testdir) / 'pycoverage'
        coverage_out_dir.mkdir(parents=True, exist_ok=False)
        coverage_all_out_files = []

    for t in tests:
        bn=os.path.basename(t)
        td=_ospathjoin(testdir,bn)
        mkdir_p(td)
        tf=open(_ospathjoin(td,'run.sh'),'w')
        tf.write('#!/usr/bin/env bash\n')
        cmdstr = bn
        if do_pycoverage:
            _whichbn = shutil.which(bn)
            _ = pathlib.Path( _whichbn ) if _whichbn else None
            if _ and _.exists() and _.read_bytes().startswith(b'#!/usr/bin/env python3'):
                coverage_outfile = coverage_out_dir / ('%s.coverage'%bn)
                coverage_all_out_files.append( coverage_outfile )
                cmdstr = 'python3 -m coverage run --data-file=%s %s'%(coverage_outfile,_whichbn)
        tf.write('mkdir rundir && cd rundir && touch ../time_start && %s &> ../output.log \n'%cmdstr)
        tf.write('EC=$?\n')
        tf.write('touch ../time_end\n')
        tf.write('echo $EC > ../ec.txt\n')
        tf.write('python3 -c \'import os.path;print(1000.0*(os.path.getmtime("../time_end")-os.path.getmtime("../time_start")))\' > ../timing_ms\n')
        if bn+'.log' in logfiles:
            ref=_ospathjoin(installdir,'tests/testref/%s.log'%bn)
            tf.write("""if [ $EC == 0 ]; then
      diff -a --ignore-all-space %s ../output.log > ../refdiff.log || diff -a --ignore-all-space -y %s ../output.log > ../refdiff_sidebyside.log; EC=$?; echo $EC > ../ecdiff.txt
    fi"""%(ref,ref))
        tf.close()

        alltests+=[bn]
        mf.write('.PHONY: %s\n\n'%bn)
        mf.write('%s:\n'%bn)
        mf.write('\t@cd ${TESTDIR}/%s && chmod +x run.sh && ./run.sh'%bn)
        mf.write('\n\n')

    mf.write('all: %s\n\n'%(' '.join(alltests)))
    mf.close()

    from .envsetup import create_install_env_clone
    test_env = create_install_env_clone()
    test_env['DISPLAY']=''
    test_env['PYTHONUNBUFFERED']='1'
    ec_global=system('make -j%i -k -f %s/Makefile all'%(njobs,testdir), env = test_env )
    rep=[]
    for t in alltests:
        td=_ospathjoin(testdir,t)
        ecdiff=None if not os.path.exists(_ospathjoin(td,'ecdiff.txt')) else int(open(_ospathjoin(td,'ecdiff.txt')).read())
        rep+=[(t,int(open(_ospathjoin(td,'ec.txt')).read()),ecdiff)]
    rep.sort()
    excerpts_to_print=[]
    if nfiltered:
        print(f'Note: {nfiltered} tests were blocked by specified filters.')
        print()
    print_header()
    for t,ec,ecdiff in rep:
        time_ms=float(open(_ospathjoin(testdir,t,'timing_ms')).read())
        ecstr='FAIL' if ec else ' OK '
        logdiffstr=' -- '
        if ecdiff is not None:
            logdiffstr='FAIL' if ecdiff else ' OK '
        info='--'
        if ec!=0 or ( (ecdiff is not None) and ecdiff):
            info=os.path.realpath(_ospathjoin(testdir,t))
            ec_global=1
            if nexcerpts>0:
                excerpts_to_print += [(t,testdir,'output.log' if ec!=0 else 'refdiff.log')]
        print("  %-37s |  %6.0f   |  %s  |   %s   | %s"%(t,time_ms,ecstr,logdiffstr,info))
        logdiffok = (ecdiff is None or not ecdiff)
        testinfo[t].update( dict(time_ms = time_ms,
                                 exitcode = ec,
                                 logfile = _ospathjoin(testdir,t,'output.log'),
                                 logdiffok = logdiffok,
                                 logdifffile = (_ospathjoin(testdir,t,'refdiff.log') if not logdiffok else None)))
    if do_pycoverage:
         _cov_infiles = ' '.join(shlex.quote(str(e.absolute().resolve())) for e in coverage_all_out_files)
         _cov_outfile = coverage_out_dir / 'combined.coverage'
         ec = system(f'cd {shlex.quote(str(coverage_out_dir))} && '
                     f'python3 -mcoverage combine --keep --data-file={shlex.quote(str(_cov_outfile.name))} {_cov_infiles}')
         if ec!=0:
             print("ERROR: py-coverage combine step failed!")
         else:
             reportdir_quoted = shlex.quote(str((coverage_out_dir/"htmlreport").absolute().resolve()))
             ec = system('python3 -mcoverage html '
                         f'--data-file={shlex.quote(str(_cov_outfile.absolute().resolve()))}'
                         f'  --directory={reportdir_quoted} --ignore-errors')
             if ec!=0:
                 print("ERROR: py-coverage html report generation failed!")
             else:
                 print(f"Pycoverage report available in: {reportdir_quoted}/index.html")

    assert len(alltests)==len(set(alltests))

    print_footer()

    do_junitxml = True
    if do_junitxml:
        testxmlwriter = testxml.TestXMLWriter()
        for testcmd in tests:
            nfo = testinfo[testcmd]
            nfo['time_ms'] = 100.0
            test = testxml.Test(testcmd,nfo['pkgname'],nfo['time_ms']*0.001)
            if nfo['exitcode'] != 0:
                test.set_run_failure(nfo['logfile'])
            elif not nfo['logdiffok']:
                test.set_reflog_failure(nfo['logdifffile'])
            testxmlwriter.add_test(test)
        outfile = pathlib.Path(testdir) / 'simplebuild_test_results_junitformat.xml'
        with outfile.open('wt') as f:
            for line in testxmlwriter.generate_xml():
                f.write('%s\n'%line)
        print()
        print(f'  Test results are also summarised in: {outfile.name}')
        print()

    if excerpts_to_print:
        for t,testdir,logname in excerpts_to_print:
            print('\n====>\n====> First %i lines from %s/%s:\n====>'%(nexcerpts,t,logname))
            system('head -%i %s'%(nexcerpts,_ospathjoin(testdir,t,logname)))
            print('====> (end of %s/%s)\n'%(t,logname))
            print('\n====>\n====> Last %i lines from %s/%s:\n====>'%(nexcerpts,t,logname))
            system('tail -%i %s'%(nexcerpts,_ospathjoin(testdir,t,logname)))
            print('====> (end of %s/%s)\n'%(t,logname))

    sys.stdout.flush()
    sys.stderr.flush()

    if ec_global:
        print('  %sERROR: Some tests failed!%s'%(col_bad,col_end))
        print()
    else:
        print('  %sAll tests completed without failures!%s'%(col_ok,col_end))
        print()
    #Used special exit code (73) to indicate test failures.
    return 0 if ec_global==0 else 73



import os
import errno
import sys
import subprocess
import shutil
import contextlib
import tempfile
import atexit

__all__ = [ 'quote', 'mkdir_p', 'chmod_x', 'rm_rf', 'rm_f', 'isemptydir',
            'system','system_throw', 'which', 'quote_cmd', 'recursive_find',
            'remove_atexit' ]

#NB: This module is still being imported by python2 in a few unit tests!!!

try:
    #Python3: quote from shlex + make pathlib-aware:
    from shlex import quote as _shlex_quote
    def quote(s):
        return _shlex_quote(str(s) if hasattr(s,'__fspath__') else s)
except ImportError:
    #Python2: quote from pipes:
    from pipes import quote

def mkdir_p(path,*args):
    """python equivalent of 'mkdir -p ...'"""
    if args:
        mkdir_p(path)
        for a in args:
            mkdir_p(a)
        return
    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

def chmod_x(path):
    """python equivalent of 'chmod +x ...'"""
    os.chmod(path, os.stat(path).st_mode | 0o111)

def rm_f(path):
    try:
        os.remove(path)#only succeeds if path exists and is not a directory
    except OSError as e:
        if e.errno == errno.ENOENT:#no such file or directory
            return#not an error, rm -f should exit silently here
        raise

def rm_rf(path):
    #First attempt with os.remove, and only in case it was a directory go for
    #shutil.rmtree (since shutil.rmtree prints error messages if fed symlinks to
    #directories):
    try:
        os.remove(path)
    except OSError as e:
        if e.errno == errno.ENOENT:#no such file or directory
            return#not an error, rm -rf should exit silently here
        elif e.errno != errno.EISDIR and e.errno!=errno.EPERM:
            raise
    #is apparently a directory
    try:
        shutil.rmtree(path)
    except OSError as e:
        if e.errno == errno.ENOENT:
            return#Sudden disappearance is still ok.
        raise

def isemptydir(path):
    try:
       for f in os.listdir(path):
           return False
    except OSError as e:
        if e.errno == errno.ENOENT:#no such file or directory
            return False
        if e.errno == errno.ENOTDIR:#not a directory
            return False
        raise
    return True

def which(cmd):
    """Function like BASH which (returns None if cmd not found)"""
    try:
        import shutil
    except ImportError:
        shutil = None
    if shutil and hasattr(shutil,'which'):
        cmd = shutil.which(cmd)
    else:
        from distutils.spawn import find_executable
        cmd=find_executable(cmd)
    if not cmd:
        return None
    return os.path.abspath(os.path.realpath(cmd))

def quote_cmd(cmd):
    if isinstance(cmd,list):
        return ' '.join(quote(c) for c in cmd)
    return cmd

def system(cmd,catch_output=False,env=None):
    """A better alternative to os.system which flushes stdout/stderr, makes sure
       the shell is always bash, and wraps exit codes larger than 127 to
       127. Set catch_output to True to instead return both exit code and the
       output of the command in a string.

    """
    #flush output, to avoid confusing ordering in log-files:
    sys.stdout.flush()
    sys.stderr.flush()
    #rather than os.system, we call "bash -c <cmd>" explicitly through
    #the subprocess module, making sure we can always use bash syntax:
    cmd=['bash','-c',cmd]
    #wrap exit code to 0..127, in case the return code is passed on to
    #sys.exit(ec):
    def fixec(ec):
        return ec if (ec>=0 and ec<=127) else 127
    if catch_output:
        try:
            p = subprocess.Popen(cmd,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,env=env)
            output = p.communicate()[0]
        except subprocess.CalledProcessError:
            #todo: in case of failure we should return the output as well!
            return 1,''
        return fixec(p.returncode),output
    else:
        ec=subprocess.call(cmd,env=env)
    sys.stdout.flush()
    sys.stderr.flush()
    return fixec(ec)

def system_throw(cmd,catch_output=False,env=None):
    """same as system except doesn't return exit code and throws RuntimeError in
    case of non-zero exit code"""
    out=None
    if catch_output:
        ec,out = system(cmd,catch_output,env=env)
    else:
        ec = system(cmd,catch_output,env=env)
    if ec:
        raise RuntimeError('command failed: %s'%cmd)
    return out

def recursive_find(searchdir,filenamepattern):
    import os
    import fnmatch
    # matches=set()
    for root, dirnames, filenames in os.walk(searchdir):
        for filename in fnmatch.filter(filenames+dirnames, filenamepattern):
            yield os.path.join(root, filename)

_files_to_clean = set()
def remove_atexit(*filelist):
    global _files_to_clean
    for f in filelist:
        _files_to_clean.add(os.path.abspath(os.path.realpath(f)))

def _filecleaner():
    global _files_to_clean
    for f in _files_to_clean:
        rm_rf(f)
atexit.register(_filecleaner)

def exec_bash_and_updateenv(cmd):
    """Execute a given bash command and update environment variables in current
    python process accordingly

    """
    import os
    import sys
    import subprocess
    import shlex
    import json
    #Execute with /usr/bin/env bash and use python3+json to stream the resulting
    #os.environ to stdout, using a well-defined encoding:
    import pathlib
    pyintrp = shlex.quote(str(pathlib.Path(sys.executable)))
    pycmd=( 'import os,sys,json;sys.stdout.buffer.write('
            'json.dumps(dict(os.environ)).encode("utf8"))')
    cmd=['/usr/bin/env','bash', '-c', '%s && %s -c %s'%(cmd,
                                                        pyintrp,
                                                        shlex.quote(pycmd))]
    pipe = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    #Unpack the streamed os.environ, using the same encoding:
    env = json.loads(pipe.stdout.read().decode('utf8'))
    os.environ = env

def source_bash_file(fn):
    """Source a given bash file and update environment variables in current
    python process accordingly"""
    import pathlib
    import shlex
    exec_bash_and_updateenv('source %s'%shlex.quote(str(pathlib.Path(fn))))

@contextlib.contextmanager
def changedir(subdir):
    """Context manager for working in a given subdir and then switching back"""
    the_cwd = os.getcwd()
    os.chdir(subdir)
    try:
        yield
    finally:
        os.chdir(the_cwd)

@contextlib.contextmanager
def work_in_tmpdir():
    """Context manager for working in a temporary directory (automatically
    created+cleaned) and then switching back"""
    the_cwd = os.getcwd()
    try:
        with tempfile.TemporaryDirectory() as tmpdir:
            os.chdir(tmpdir)
            yield
    finally:
        os.chdir(the_cwd)



import os
from pathlib import Path
from . import dirs
from . import langs
from . import conf
from . import error
join=os.path.join

def parse_depfile(pkgdir):
    filename=os.path.join(pkgdir,conf.package_cfg_file)
    pkg_name = os.path.basename(pkgdir)
    def _err(m=''):
        error.error('Package "%s" has invalid %s file%s'%(pkg_name,conf.package_cfg_file,m))
    with open(filename,'rt') as fh:
        for iline1,ll in enumerate(fh):
            ll=ll.split('#')[0].strip()
            if ll.startswith('package(') and ll.endswith(')'):
                ll=ll[8:-1]
                extdeps=[]
                pkgdeps=[]
                extra_cflags=[]
                extra_ldflags=[]
                extra_incdeps=[]
                current=None
                for e in ll.split():
                    if e=='USEPKG':
                        current=pkgdeps
                    elif e=='USEEXT':
                        current=extdeps
                    elif e=='EXTRA_COMPILE_FLAGS':
                        current=extra_cflags
                    elif e=='EXTRA_LINK_FLAGS':
                        current=extra_ldflags
                    elif e=='EXTRA_INCDEPS':
                        current=extra_incdeps
                    else:
                        if current is None:
                            _err()
                        current += [e]
                return ( set(extdeps),
                         set(pkgdeps),
                         extra_cflags,
                         extra_ldflags,
                         extra_incdeps )
        _err(' : missing package() statement.')

def _check_case_insensitive_duplication(path_str):
    path = Path(path_str)
    if not path.exists():
        return
    parentContent = [d.lower() for d in os.listdir(path.parent)]
    occurance = parentContent.count(path.name.lower())
    if not occurance == 1:
        from . import error
        error.error('Directory (and file) names differing only in casing are'
                    ' not allowed, due to being a potential source of error'
                    ' for different file systems. \nProblem occured with %s'%(path_str))

def find_pkg_dirs( basedir_list ):
    pkgdirs={}
    from .pkgutils import _find_pkg_dirs_under_basedir as _findpkgs
    for basedir in basedir_list:
      _check_case_insensitive_duplication(basedir)
      tmp_dirs = _findpkgs( basedir,
                            cfgfilename = conf.package_cfg_file )
      if tmp_dirs:
        pkgdirs.update({d:basedir for d in tmp_dirs})
      elif not basedir==dirs.projdir:
        error.error("No packages found in %s!"%basedir)
    return pkgdirs

class Package:
    #Only construct this with the PackageLoader

    def __lt__(self,o):
        assert self.name != o.name
        return self.name < o.name

    def __init__(self,basedir,reldirname,enabled):
        self.dirname = os.path.join(basedir,reldirname)
        self.dirpath = basedir.joinpath(reldirname)
        self.__haslib=None
        self.reldirname = reldirname
        #self.reldirnamewithparent = os.path.relpath(dirname,basedir+'/..')
        self.name = os.path.basename(reldirname)
        assert isinstance(self.name,str)
        assert isinstance(self.dirname,str)#fixme
        assert isinstance(self.reldirname,str)#fixme
        self.enabled=enabled
        self.is_setup=False
        self._any_parent_changed=False
        self.__deps_names=None
        self.__incs_updated=False
        self.__runnables=set()#to guard against clashes from scripts and app_, etc.
        self.__files_in_pkg_changed = False

    def register_runnable(self,n):
        if n in self.__runnables:
            error.error('Package "%s" provides multiple sources for the command "%s"'%(self.name,n))
        self.__runnables.add(n)

    def update_incs(self):
        if not self.__incs_updated:
            self.__incs_updated = True
            d=dirs.pkg_dir(self,'libinc')
            from . import db
            db.db['pkg2inc'][self.name]= langs.headers_in_dir(d) if os.path.isdir(d) else set()

    def setup(self,name2object,autodeps,*,enable_and_recurse_to_deps=False):
        #Ensures cfg file is parsed and direct dep links are established.
        #Recursively setups (and enables) deps if requested.
        if self.is_setup:
            return
        self.is_setup=True
        pcf=join(self.dirname,conf.package_cfg_file)
        ed,pd,cf,lf,extra_incdeps=parse_depfile(self.dirname)
        if self.name not in autodeps:
            pd.update(autodeps)
        self.direct_deps_extnames = ed
        self.direct_deps_pkgnames = pd
        self.direct_clients=set()
        self.extraflags_comp = cf
        self.extraflags_link = lf
        self.extra_include_deps = []
        for eid in extra_incdeps:
            eid=[e.strip() for e in eid.split(':')]
            if len(eid)!=2 or not eid[0] or not eid[1]:
                error.error('Malformed EXTRA_INCDEPS statement in %s'%pcf)
            self.extra_include_deps += [eid]

        ll=[]
        for n in pd:
            o=name2object(n)
            if not o:
                if n in autodeps:
                    error.error('Unknown package "%s" specified as auto dependency'%n)
                else:
                    error.error('Unknown package "%s" specified in %s'%(n,pcf))
            ll+=[o]
        self.direct_deps=ll
        if enable_and_recurse_to_deps:
            self.enabled=True
            for p in ll:
                p.setup(name2object,autodeps,enable_and_recurse_to_deps=True)
        self.__deps=None
        self.__extdeps=None
        self.__deplock=False

    def deps(self):
        #all pkgs we depend on, directly or indirectly.
        if self.__deps is not None:
            return self.__deps
        if self.__deplock:
            if self in self.direct_deps:
                error.error('Package "%s" depends on itself'%self.name)
            else:
                error.error('Detected circular dependency involving package "%s"'%self.name)
        self.__deplock=True
        d=set(self.direct_deps)
        for p in self.direct_deps:
            d.update(p.deps())
        self.__deplock=False
        self.__deps=d
        return d

    def deps_names(self):
        #names all pkgs we depend on, directly or indirectly, including ourselves.
        if self.__deps_names is not None:
            return self.__deps_names
        self.__deps_names = set(p.name for p in self.deps())
        return self.__deps_names

    def extdeps(self):
        #all extdeps we depend on, directly or indirectly.
        if self.__extdeps is not None:
            return self.__extdeps
        self.__extdeps=set(self.direct_deps_extnames)
        for p in self.direct_deps:
            self.__extdeps.update(p.extdeps())
        return self.__extdeps

    def disable(self):
        #Disable direct clients (only if we have actually been setup, otherwise we don't have direct_clients
        self.enabled=False
        _dc = self.direct_clients if hasattr(self,'direct_clients') else []
        for dc in _dc:
            dc.disable()

    def set_files_in_pkg_changed(self):
        self.__files_in_pkg_changed = True
        self.__set_parent_changed_recursively()
        self.__set_files_in_pkg_changed_recursively()

    def files_in_pkg_changed(self):
        return self.__files_in_pkg_changed

    def any_parent_changed(self):
        return self._any_parent_changed

    def __set_parent_changed_recursively(self):
        self._any_parent_changed=True
        for dc in self.direct_clients:
            dc.__set_parent_changed_recursively()

    def __set_files_in_pkg_changed_recursively(self):
        self.__files_in_pkg_changed=True
        for dc in self.direct_clients:
            dc.__set_files_in_pkg_changed_recursively()

    def has_lib(self):
        if self.__haslib is None:
            self.__haslib = os.path.isdir(dirs.pkg_dir(self,'libsrc'))
        return self.__haslib

    def all_clients(self):
        ic=set()
        ic.update(self.direct_clients)
        for c in self.direct_clients:
            ic.update(c.all_clients())
        return ic

    def info_as_dict(self):
        """Returns info as dictionary, containing not custom classes and suitable for
           pickling. Should be called after setup and after targets have been built, in
            order to guarantee complete info"""
        d = { 'name':self.name,
              'dirname':self.dirname,
              'reldirname':self.reldirname,
              'enabled' : self.enabled }
        if not self.enabled:
            return d
        from . import db
        d.update({ 'deps_pkgs' : self.deps_names(),
                   'deps_pkgs_direct' : self.direct_deps_pkgnames,
                   'deps_ext' : self.extdeps(),
                   'deps_ext_direct' : self.direct_deps_extnames,
                   'extraflags_comp' : self.extraflags_comp,
                   'extraflags_link' : self.extraflags_link,
                   'extra_include_deps' : self.extra_include_deps,
                   'has_lib' : self.has_lib(),
                   'clients' : [p.name for p in sorted(self.all_clients())],
                   'clients_direct' : [p.name for p in sorted(self.direct_clients)],
                   'runnables':db.db['pkg2runnables'].get(self.name,set()),
                   'reflogs':db.db['pkg2reflogs'].get(self.name,set()),
        })
        return d

    def dumpinfo(self,autodeps):
        from . import col
        from . import env
        width=max(75,len(self.dirname)+30)
        from .io import print, print_prefix
        def _format(ll):
            if not ll:
                return col.darkgrey+'<none>'+col.end
            from . import formatlist
            return '\n'.join(
                formatlist.formatlist(
                    [(n,col.ok if b else col.bad) for n,b in sorted(ll)],
                    width-40,
                    indent_first='',
                    indent_others=print_prefix+' '*27)
            )
        extdeps_direct = [(e,env.env['extdeps'][e]['present']) for e in self.direct_deps_extnames]
        extdeps_indirect = [(e,env.env['extdeps'][e]['present']) for e in self.extdeps() if e not in self.direct_deps_extnames]
        def nameformat(n):
            return n+'*' if n in autodeps else n
        clients_direct=[(nameformat(p.name),p.enabled) for p in self.direct_clients]
        clients_indirect=[(nameformat(p.name),p.enabled) for p in self.all_clients().difference(self.direct_clients)]
        deps_direct = [(nameformat(p.name),p.enabled) for p in self.direct_deps]
        deps_indirect = [(nameformat(p.name),p.enabled) for p in self.deps() if p not in self.direct_deps]

        s='='*int((width-len('Package information')-2)//2)
        l1='%s Package information %s'%(s,s)
        print(l1)
        print('Package name            : %s'%(_format([(self.name,self.enabled)])))
        print('Enabled                 : %s'%(_format([('Yes' if self.enabled else 'No',self.enabled)])))
        print('Relative location       : %s%s%s'%(col.blue,self.reldirname,col.end))
        print('Physical location       : %s%s%s'%(col.yellow,self.dirname,col.end))
        print('Direct relations')
        print('  Dependencies          : %s'%(_format(deps_direct)))
        print('  Clients               : %s'%(_format(clients_direct)))
        print('  External dependencies : %s'%(_format(extdeps_direct)))
        print('Indirect relations')
        print('  Dependencies          : %s'%(_format(deps_indirect)))
        print('  Clients               : %s'%(_format(clients_indirect)))
        print('  External dependencies : %s'%(_format(extdeps_indirect)))
        print('='*len(l1))

class PackageLoader:

    def __init__(self,
                 pkgdirs,
                 pkg_filter_obj=None,
                 autodeps=None,
                 load_all=False):
        """Locates package directories and inits wrapper Package objects for
        each.  It will respect select_pkg_filter and attempt to be economical
        about loading the package configuration files. Filters will receive
        package reldirname and basename as input.'
        """
        #use load_all to (inefficiently) load all cfg info even when filter is
        #set (for queries about pkg metadata)

        #pkg_filter_obj.dump()
        if pkg_filter_obj.fully_open():
            pkg_filter_obj = None

        #Note, when select_pkg_filter is set, we do not need to load ALL cfg
        #files, hence the difference in treatment below.
        self.autodeps=set(autodeps if autodeps else [])

        #1) Sanity check arguments:
        #Commented for now, since we want to enable both NOT= and ONLY=
        #variables at the same time. Leaving here in case we encounter
        #unexpected side-effects:
        #if select_pkg_filter and exclude_pkg_filter:
        #    error.error("Can't exclude packages when simultaneously "
        #                "requesting to enable only certain packages")

        #2) Construct Package objects and name->object maps:
        default_enabled = pkg_filter_obj is None

        n2p={}
        lowercased_pkgs=set()
        pkgs=[]
        from .pkgutils import pkgname_valid_or_errmsg
        for pd,basedir in pkgdirs.items():
            pkg_reldir = os.path.relpath(pd,basedir)
            pkg_name = os.path.basename(pd)
            assert len(pkg_name)>=3
            p=Package(basedir,pkg_reldir,default_enabled)
            pkgs+=[p]
            ln=p.name.lower()
            if p.name in n2p:
                error.error('Duplicate package name: "%s"'%p.name)
            if ln in lowercased_pkgs:
                clashes='" "'.join(p.name for p in pkgs if p.name.lower()==ln)
                error.error('Package names clash when'
                            ' lowercased: "%s" '%clashes)
            lowercased_pkgs.add(ln)
            pkgnameerrmsg = pkgname_valid_or_errmsg(p.name)
            if pkgnameerrmsg is not None:
                error.error(pkgnameerrmsg)
            n2p[p.name]=p
        self.pkgs=pkgs
        self.name2pkg=n2p

        #3) Load contents of all or some of the package config files.
        g=self.name2pkg.get
        def pkg_name2obj(pn):
            return g(pn,None)
        if pkg_filter_obj:
            for p in pkgs:
                if pkg_filter_obj.passes( p.name, p.reldirname ):
                    p.setup(pkg_name2obj,
                            autodeps,enable_and_recurse_to_deps=True)
        if load_all or not pkg_filter_obj:
            for p in pkgs:
                p.setup(pkg_name2obj,autodeps)

        #4) setup client links (only to clients which are setup):
        for p in pkgs:
            if p.is_setup:
                for pd in p.direct_deps:
                    if pd.is_setup:
                        pd.direct_clients.add(p)

        for p in self.enabled_pkgs_iter():
            p.deps()#must setup deps of all enabled packages, otherwise we can't be sure to detect circular deps

    def check_no_forbidden_extdeps( self, possible_extdeps ):
        for p in self.pkgs:
            if not hasattr(p,'direct_deps_extnames'):
                continue#pkg.info was never parsed (possibly the package was disabled by filter => do not load just to validate)
            forbidden = p.direct_deps_extnames - possible_extdeps
            forbidden = forbidden.pop() if forbidden else None
            if forbidden:
                error.error('Unknown external dependency "%s" specified in %s/%s'%(forbidden,p.dirname,conf.package_cfg_file)+
                          '\Available values in this configuration are: "%s"'%'", "'.join(possible_extdeps))

    def enabled_pkgs_iter(self):
        for p in self.pkgs:
            if p.enabled:
                yield p

    def active_deps_extnames(self):
        s=set()
        for e in (p.direct_deps_extnames for p in self.enabled_pkgs_iter()):
            s.update(e)
        return s

    def disable_pkgs_with_extdeps(self,extdeps):
        if not extdeps:
            return#nothing to do
        prev_enabled=set(self.enabled_pkgs_iter())
        for p in self.pkgs:
            if p.enabled and not extdeps.isdisjoint(p.direct_deps_extnames):
                p.disable()
        return prev_enabled.difference(set(self.enabled_pkgs_iter()))



_exported_cmds = []

def _eval_makefile_vars( s, varmap ):
    if '${' not in s:
        return s
    any_hits = False
    for k,v in varmap:
        _k = '${%s}'%k
        if _k in s:
            any_hits = True
            s = s.replace(_k,v)
    return _eval_makefile_vars(s, varmap) if any_hits else s

def export_pkg( pkg, varmap ):
    import pathlib
    #Trying to follow https://clang.llvm.org/docs/JSONCompilationDatabase.html
    #and the format used by CMake's -DCMAKE_EXPORT_COMPILE_COMMANDS=ON
    from . import dirs
    wd = str(dirs.makefiledir)
    for t in pkg.targets:
        if hasattr(t,'export_compilation_commands'):
            for cmd, sourcefile in t.export_compilation_commands():
                sf = pathlib.Path(_eval_makefile_vars(sourcefile,varmap))
                if not sf.exists():
                    from . import error
                    error.error('Exported json contains non-existing source'
                                ' file: %s'%sourcefile)
                _exported_cmds.append( {
                    'directory' : wd,
                    'command' : _eval_makefile_vars(cmd,varmap),
                    'file' : str(sf.absolute().resolve())
                } )

def finalise( destfile ):
    import json
    from .io import print
    print(f"Exporting {len(_exported_cmds)} compilation"
          f" commands to file: {destfile.absolute()}")
    destfile.write_text( json.dumps( _exported_cmds, indent='  ' ) )



class PkgFilter:

    """Parse and implement the package filters defined in simplebuild.cfg data."""

    # If any positive filters are defined, a pkg must pass at least one of them to pass.
    # If any negative filters are defined, a pkg must NOT pass any of them, to pass.
    # NB: The same logic is applied to sb --testfilters

    #This is documented on in doc/source/simplebuilddotcfg.rst

    def __init__( self, filter_definition_list ):
        import fnmatch
        import re
        # expand on commas, strip and prune empty filters, remove duplicates:
        expanded_list = []
        for e in filter_definition_list:
            expanded_list += e.split(',')
        final_list = []
        for e in expanded_list:
            e = e.strip()
            is_rejection = e.startswith('!')
            if is_rejection:
                e = e[1:]
            if not e:
                #Silently ignore empty entries (both "!" and "")
                continue
            e = ( is_rejection, e )
            if e not in final_list:
                final_list.append( e )
        final_list.sort()

        def fmt( is_rejection, e ):
            return f'!{e}' if is_rejection else e
        single_str_list = []
        acceptance_filters = []
        rejection_filters = []
        for is_rejection, f in final_list:
            forig = fmt(is_rejection,f)
            single_str_list.append( forig )
            is_abs_dir_pattern = False
            if f.startswith('RE::'):
                onreldir = ''
                regex = f[4:]
            else:
                is_abs_dir_pattern = f.startswith('/')
                if is_abs_dir_pattern:
                    from . import error
                    error.error('Absolute paths not supported in pkg filters')
                looks_like_reldir = '/' in f
                onreldir = './' if looks_like_reldir else None
                if onreldir and not f.startswith('./'):
                    f = './'+f
                regex = fnmatch.translate( f )
            try:
                reobj = re.compile(regex)
            except re.error as re:
                from . import error
                error.error(f'Invalid syntax in pkg filter pattern: "{forig}"')
            if is_rejection:
                rejection_filters.append( ( onreldir, reobj ) )
            else:
                acceptance_filters.append( ( onreldir, reobj ) )
        self.__accept_filters = tuple(acceptance_filters)
        self.__reject_filters = tuple(rejection_filters)
        self.__as_single_str = ','.join(single_str_list)

    def fully_open( self ):
        return not self.__accept_filters and not self.__reject_filters

    def dump( self ):
        from .io import print
        print( "PkgFilter::" )
        for onreldir, reobj in self.__accept_filters:
            print( f'  -> ACCEPTANCE Pattern="{reobj.pattern}", applied_to={f"{onreldir}reldir" if onreldir is not None else "pkgname"}')
        for onreldir, reobj in self.__reject_filters:
            print( f'  -> REJECTION Pattern="{reobj.pattern}", applied_to={f"{onreldir}reldir" if onreldir is not None else "pkgname"}')

    def passes( self, pkg_name, pkg_reldir ):
        if self.__accept_filters:
            accepted = False
            for onreldir, reobj in self.__accept_filters:
                p = onreldir+pkg_reldir if onreldir is not None else pkg_name
                if reobj.match( p ) is not None:
                    accepted = True
                    break
            if not accepted:
                return False
        for onreldir, reobj in self.__reject_filters:
            p = onreldir+pkg_reldir if onreldir is not None else pkg_name
            if reobj.match( p ) is not None:
                return False
        return True

    def apply( self, pkg_names_and_reldirs ):
        for name, reldir in pkg_names_and_reldirs:
            if self.passes( name, reldir ):
                yield name

    def as_string( self ):
        return self.__as_single_str

    def __str__( self ):
        return self.as_string()



from . import utils
from . import langs

def create_include_decoder(exts):
    #decodes c++ include statements to look for possible includes to private header files (in same dir) or in libinc of own or other package.
    #
    #The decoder Returns (pkg,filename).
    #filename==None means not a valid match
    #pkg==None means possible private include, pkg!=None means possible include from that pkg
    import re
    exts = [(e.encode('ascii') if hasattr(e,'encode') else e) for e in exts]
    pattern = b'^\s*#\s*include\s*"\s*(([a-zA-Z0-9_]+/)?([a-zA-Z0-9_]+){1}(%s))\s*"'%(b'|'.join(exts))
    #match=re.compile(pattern.encode('ascii') if hasattr(pattern,'encode') else pattern).match
    match=re.compile(pattern).match
    def decoder(ll):
        m=match(ll)
        if not m:
            return None,None
        _,pkg,fn,ext = m.groups()
        fn=b'%s%s'%(fn,ext)
        if pkg:
            pkg=pkg[:-1]
        return pkg,fn
    return decoder
include_decoder = create_include_decoder(langs.hdrext2lang.keys())

def find_includes(cfile,pkg):
    #We only look for includes which might be to files in same dir or other
    #packages, i.e. those using "../.." or "..", not <> or "../../.."
    #
    #We might find a few too many if there are ifdefs or /*..*/  style comments.
    #
    #For efficiency, initial dig through file use grep command:
    ec,output=utils.run(['grep','.*#.*include.*"..*"',cfile])
    if ec!=0:
        return None,None
    possible_privincs=set()
    possible_pkgincs=set()
    def bytes2str( b ):
        return b.decode('ascii')
    for ll in output.splitlines():
        pkgname,fn = include_decoder(ll)
        if fn:
            if pkgname:
                possible_pkgincs.add((bytes2str(pkgname),bytes2str(fn)))
            else:
                possible_privincs.add(bytes2str(fn))
    if pkg.extra_include_deps:
        from os.path import relpath
        rp=relpath(cfile,pkg.dirname)
        for fn0,incdep in pkg.extra_include_deps:
            if fn0==rp:
                pkgname,fn = include_decoder(
                    b'#include "%s"'%(incdep.encode('ascii')
                                      if hasattr(incdep,'encode') else incdep)
                )
                if fn:
                    if pkgname:
                        possible_pkgincs.add((bytes2str(pkgname),bytes2str(fn)))
                    else:
                        possible_privincs.add(bytes2str(fn))
    return possible_privincs,possible_pkgincs



"""

(obsolete - use 'FindData3') Python module providing access to data files of the
various packages

"""

import os
import sys

#We use a little class/sys.modules trick to allow both: Core.FindData(..) and
#Core.FindData.findData(..)

class _themodule:
    _os = os#this workaround is needed for some weird python2-lxplus
            #bug (DGSW-466)
    def findData(self,pkg,filename=None):
        if filename is None and not pkg.startswith('/'):
            fs=pkg.split('/')
            if len(fs)==2:
                ret = self.findData(*fs)
                if self._os.path.exists(ret):
                    return ret
            ret = pkg
        else:
            ret = self._os.path.join( self._os.environ['SBLD_DATA_DIR'],
                                      pkg,
                                      filename )
        return ret if self._os.path.exists(ret) else ""
    def __call__(self,pkg,filename=None):
        return self.findData(pkg,filename)
sys.modules[__name__] = _themodule()



import os
from . import dirs
from . import utils

#Database for remembering package info between invocations.

db = { 'pkg2timestamp' : {},#to detect when files changed
       'pkg2inc' : {},#keep track of provided public headers (populate at target __init__ stage)
       'pkg2parts' : {},#keep track of targets to the point where we can cleanup the install area when user removes files or dirs
       'pkg2reldir' : {},#keep track of pkg locations
       'pkg2runnables' : {},#Keep track of runnables (with cmdname, e.g. <pn>_bla_lala)
       'pkg2reflogs' : {},#Keep track of reference logs (e.g. <pn>_bla_lala.log)
       'pkg2directextdeps' : {},#list of extdep's directly used by package.
   }

#initial load:
dbfile=dirs.blddir / 'globdb.pkl'

if os.path.exists(dbfile):#fixme: utils should have pkl_load safe which returns None if file does not exists (avoids potential race conds)
    db = utils.pkl_load(dbfile)

def save_to_file():
    utils.pkl_dump(db,dbfile)

def clear_pkg(pkgname):
    for k,v in db.items():
        if k.startswith('pkg2') and pkgname in v:
            del v[pkgname]



# We want to create XML in "JUNIT format". Apart from a few xml escaping
# functions, we do this manually since it is almost trivial.

import pathlib
from xml.sax.saxutils import escape as xml_escape_general_data
from xml.sax.saxutils import quoteattr as xml_escape_attribute_data

class Test:

    def __init__( self, cmdname, pkgname, time_elapsed_seconds ):
        self.__cmdname = cmdname
        self.__pkgname = pkgname
        self.__time_elapsed_seconds = time_elapsed_seconds
        self.__failure = None

    def time_elapsed_seconds( self ):
        return self.__time_elapsed_seconds

    def is_failure( self ):
        return self.__failure is not None

    def set_run_failure( self, logfile ):
        self.__failure = ( 'runfail', logfile )

    def set_reflog_failure( self, logdiff_file ):
        self.__failure = ( 'reflogfail', logdiff_file )

    def __readsafe( self, fn):
        s = pathlib.Path(fn).read_bytes()
        s = s.decode('utf-8','backslashreplace')
        return s if len(s)<11000 else (s[0:5000]+"\n<<<skipping to the ending>>>\n"+ s[-5000:] +"\n<<<output exceeds limit for inclusion here>>>")

    def _generate_xml(self):
        metadata = f'name={xml_escape_attribute_data(self.__cmdname)} {_fmt_time_attribute(self.__time_elapsed_seconds)} classname={xml_escape_attribute_data(self.__pkgname)}'
        if self.__failure is None:
            yield f'    <testcase {metadata} />'
        else:
            s = self.__readsafe(self.__failure[1])
            yield f'    <testcase {metadata}>'
            if self.__failure[0]=='runfail':
                s = s or '<Command gave no output>'
                msg = 'Execution failed'
            else:
                s = s or '<Diff mysteriously missing>;'
                msg = 'Output differs from reference'
            for line in f'      <failure type="failure" message={xml_escape_attribute_data(msg)}>{xml_escape_general_data(s)}</failure>'.splitlines():
                yield line
            yield '    </testcase>'

class TestXMLWriter:

    def __init__(self):
        self.__tests = []

    def add_test( self, test ):
        assert isinstance( test, Test )
        self.__tests.append( test )

    def generate_xml(self):
        ntests = len(self.__tests)
        nfailures = sum(1 for t in self.__tests if t.is_failure())
        time_sum = float(sum(t.time_elapsed_seconds() for t in self.__tests))

        yield f'<testsuites failures="{nfailures}" errors="0" tests="{ntests}" disabled="0" {_fmt_time_attribute(time_sum)}>'
        yield f'  <testsuite name="simplebuild-tests" disabled="0" failures="{nfailures}" errors="0" skipped="0" time="{time_sum}" tests="{ntests}">'
        for t in self.__tests:
            for line in t._generate_xml():
                yield line
        yield '  </testsuite>'
        yield '</testsuites>'

def _fmt_time_attribute( time_seconds ):
    return 'time=%s'%xml_escape_attribute_data('%g'%time_seconds)



# To keep things clean and allow for future migrations to different cfg systems,
# this file should be the ONLY place we query environment variables!!

def _query(n, *, boolean=False):
    import os
    _ = os.environ.get(n)
    if boolean:
        return (_ or '').lower() in ['true', 'on', 'yes', '1']
    return _

def _build_cfg():
    from .cfglocate import locate_master_cfg_file
    from .cfgbuilder import CfgBuilder
    from .singlecfg import SingleCfg
    from .pkgfilter import PkgFilter
    master_cfg_file = locate_master_cfg_file()

    if not master_cfg_file or not master_cfg_file.is_file():
        from . import error
        error.error('In order to continue, please step into a directory'
                    ' tree with a simplebuild.cfg file at its root.')

    assert master_cfg_file.is_file()

    master_cfg = SingleCfg.create_from_toml_file( master_cfg_file )
    cfg = CfgBuilder( master_cfg, master_cfg_file )
    pkgfilterobj = PkgFilter( cfg.build_pkg_filter )

    cachedir_postfix = ( '' if master_cfg.build_mode=='release'
                           else f'_{master_cfg.build_mode}' )

    import shlex
    _cmake_args = shlex.split( _query('CMAKE_ARGS') or '' )

    if master_cfg.build_mode=='release':
        _cmake_args.append('-DCMAKE_BUILD_TYPE=Release')
    else:
        assert master_cfg.build_mode=='debug'
        _cmake_args.append('-DCMAKE_BUILD_TYPE=Debug')
    _build_mode_summary_string = master_cfg.build_mode.capitalize()

    class EnvCfg:

        #These are the basic ones:
        build_dir_resolved = cfg.build_cachedir / f'bld{cachedir_postfix}'
        install_dir_resolved = cfg.build_cachedir / f'install{cachedir_postfix}'

        projects_dir = master_cfg.project_pkg_root #FIXME: Is this ok?
        extra_pkg_path = ':'.join(str(e) for e in cfg.pkg_path)#fixme: keep at Path objects.
        extra_pkg_path_list = cfg.pkg_path#New style!
        enable_projects_pkg_selection_flag = False#fixme: we could allow this?
        pkg_filter = pkgfilterobj#New style!

        #These are used in the context of conda installs:
        conda_prefix =  _query('CONDA_PREFIX')
        cmake_args =  _cmake_args

        #These are most likely almost never used by anyone (thus keeping as env
        #vars for now!):
        color_fix_code = _query('SIMPLEBUILD_COLOR_FIX')
        allow_sys_dev =  _query('SIMPLEBUILD_ALLOWSYSDEV',boolean=True)

        # NOTE: backend.py also checks environment variables to check if
        # something changed needing an automatic cmake reconf. We provide a good
        # base list here:

        reconf_env_vars = [
            #All of the above except SIMPLEBUILD_ALLOWSYSDEV:
            'PATH','SIMPLEBUILD_COLOR_FIX','CONDA_PREFIX','CMAKE_ARGS','PYTHONPATH',
            #Also this one of course:
            'SIMPLEBUILD_CFG',
        ]

        env_paths = cfg.env_paths

        build_mode_summary_string = _build_mode_summary_string

    return EnvCfg()

var = _build_cfg()



def extract_filelist( fn ):
    from . import utils
    import os
    try:
        files,srcdir,destdir = utils.pkl_load(fn)
    except EOFError:
        print("ERROR: pickle file %s ended unexpectedly."%fn)
        raise
    return set( (destdir, fnlink, os.path.join( srcdir, fnsrc ) )
                for fnsrc, fnlink in files )
def main():
    from . import utils
    import sys
    import os
    fn=sys.argv[1]
    verbose=int(sys.argv[2]) if len(sys.argv)>=3 else 0
    #verbose flag: -1 quiet, 0 normal, 1 verbose.
    do_print = verbose > 0
    newlist = extract_filelist( fn )
    oldfn = '%s.old'%fn
    oldlist = extract_filelist( oldfn ) if os.path.exists(oldfn) else set()
    utils.update_symlinks( oldlist, newlist, do_print )

if __name__=='__main__':
    import sys
    try:
        main()
    except KeyboardInterrupt:
        from .io import print
        print("<<symlink installation interrupted by user!>>>")
        #Fixme: Any way to recover state?? (if so we should also catch other errors)
        sys.stdout.flush()
        sys.stderr.flush()
        import time
        time.sleep(0.2)
        sys.exit(1)



#!/usr/bin/env python3

__author__ = 'Thomas Kittelmann'

import argparse
import pathlib
import os
import os.path
import shlex
import sys
import subprocess
import json
import contextlib

def parse_args( argv = None ):
    descr = """Use to extract information about C/C++ compiler and linker via
               CMake (with "Unix Makefiles" generator). Will respect and apply
               CMAKE_ARGS environment variable if set (used by conda-forge).

               Run without --findpackageargs and --deptargets results to get the
               basic compilation and link commands and flags.

               To help with escaping troubles, any "@@" in arguments to string
               options will be replaced with a single space.

               Any arguments following a trailing "--" will simply be forwarded
               to the CMake invocation.
    """
    if not argv:
        argv = sys.argv[:]

    parser = argparse.ArgumentParser( prog = os.path.basename(sys.argv[0]),
                                      description=descr )
    parser.add_argument( '-v', '--verbose', action='store_true', default=False )
    parser.add_argument( '-f', '--force', action='store_true', default=False )
    parser.add_argument( '-o', '--output', type=str, default=None,
                         help=("File in which to store results (in JSON format)."
                               " Use 'stdout' (default) and 'stdout_json' to get "
                               "the result pretty printed instead of stored.") )
    parser.add_argument( '-l', '--language', choices=('C','CXX'), default='CXX',
                         help='Language' )
    parser.add_argument( '--findpackage', type=str, nargs='?',const='',
                         help='Arguments for find_package(..) in CMake' )
    parser.add_argument( '--deptargets', type=str, nargs='?', const='',
                         help=('CMake targets/libraries (from the find_package'
                               ' call) which should be investigated for additional'
                               ' compilation and link flags') )

    parser.add_argument( '--respectcmakeargs', action='store_true', default=False,
                         help = ("Respect the CMAKE_ARGS environment variable"
                                 " associated with conda-forge environments." ) )
    parser.add_argument( '--cmakecommand', type=str, default=None,
                         help=("Path to cmake command. If not specified, the"
                               " 'cmake' command in the current path will be used" ) )

    args = argv[1:]
    cmake_args = []
    if '--' in args:
        cmake_args = args[args.index('--')+1:]
        args = args[0:args.index('--')]

    args = parser.parse_args( args )
    for a in ('findpackage','deptargets'):
        v = getattr(args,a)
        if v is not None and not v:
            parser.error(f'Missing value for --{a}')
    if ( args.findpackage is None ) != ( args.deptargets is None ):
        parser.error('Must provide both or neither of the flags'
                     ' --findpackage and --deptargets')
    for a in ('findpackage','deptargets'):
        v = getattr(args,a)
        if v and '@@' in v:
            setattr(args,a,v.replace('@@',' '))
    if not args.output:
        args.output = 'stdout'
    elif args.output not in ('stdout','stdout_json'):
        p = pathlib.Path(args.output)
        p.expanduser()
        p = p.absolute().resolve()
        if p.is_dir():
            parser.error(f'Error: Output file {p} is a directory.')
        if p.exists() and not args.force:
            parser.error(f'Error: Output file {p} already'
                         ' exists (specify --force to overwrite).')
        if not p.parent or not p.parent.is_dir():
            parser.error('Error: Invalid or missing directory'
                         f' for output file {p}.')
        args.output = p

    setattr( args, 'cmakeargs', cmake_args )
    return args

def get_cmake_command( args ):
    if args.cmakecommand:
        ll = [ args.cmakecommand ]
    else:
        import shutil
        ll = [ shutil.which('cmake') ]
    if args.respectcmakeargs:
        #for conda-forge we should always apply CMAKE_ARGS:
        ll += shlex.split(os.environ.get('CMAKE_ARGS',''))
    if args.cmakeargs:
        ll += args.cmakeargs
    #Our procedure only works with the Unix Makefiles generator:
    ll += ['-G','Unix Makefiles']
    return ll

class ParseCmd:
    def __init__( self, cmd, input_file ):
        ll = shlex.split(cmd)
        assert len(ll)>=1
        assert ll.count('-o')==1
        _ = []
        idx_o = ll.index('-o')
        for i,e in enumerate(ll):
            if idx_o <= i <= idx_o+1:
                _.append(e)
                continue
            if e.endswith(input_file) or e.endswith(input_file+'.o'):
                _.append('THE-INPUT-FILE-HERE')
            else:
                _.append(e)
        ll = _

        self.progname = ll[0]
        self.args_preoutput = ll[1:idx_o]
        self.args_postoutput = ll[idx_o+2:]
        self.outputname = ll[idx_o]

    def all_args(self):
        return self.args_preoutput + self.args_postoutput

    def __str__(self):
        return ' '.join(shlex.quote(e) for e in ( [ self.progname ]
                                                  + self.args_preoutput
                                                  + ['-o','THE-OUTPUT-FILE-HERE']
                                                  + self.args_postoutput) )

    def extra_args( self, basecmd ):
        l1 = self.all_args()[:]
        l0 = basecmd.all_args()[:]
        ll = []
        for e in l1:
            if l0 and e==l0[0]:
                l0.pop(0)
                continue
            ll.append(e)
        return ll

def create_cmakeliststxt( args ):
    dummy_usage_statements = ''
    if args.respectcmakeargs:
        #Something in conda-forges CMAKE_ARGS might trigger an unused variable
        #warning, so we make sure we explicitly use them with a dummy statement:
        for e in shlex.split(os.environ.get('CMAKE_ARGS','')):
            if e.startswith('-D'):
                e = e[2:].split('=')[0]
                dummy_usage_statements += '\n# Avoid warning if %s is not used:\n'%e
                dummy_usage_statements += 'if ( DEFINED "%s" )\n'%e
                dummy_usage_statements += '  set( tmp "${%s}" )\n'%e
                dummy_usage_statements += 'endif()\n'
    cmake = """cmake_minimum_required(VERSION 3.24.2...3.27.6)
project( dummy LANGUAGES %s )
%s
set_source_files_properties( "${CMAKE_CURRENT_SOURCE_DIR}/dummy_nodep.input"
                             PROPERTIES LANGUAGE %s)
add_executable(dummy_nodep_app ${CMAKE_CURRENT_SOURCE_DIR}/dummy_nodep.input )
set_target_properties( dummy_nodep_app PROPERTIES LANGUAGE %s )
set_target_properties( dummy_nodep_app PROPERTIES EXPORT_COMPILE_COMMANDS ON )
"""%(args.language,dummy_usage_statements,args.language,args.language)

    if args.findpackage:
        fpargs = args.findpackage
        if 'REQUIRED' not in fpargs:
            fpargs+=' REQUIRED'
        cmake += """
find_package(%s)
set_source_files_properties( "${CMAKE_CURRENT_SOURCE_DIR}/dummy_withdep.input"
                             PROPERTIES LANGUAGE %s)
add_executable(dummy_withdep_app ${CMAKE_CURRENT_SOURCE_DIR}/dummy_withdep.input )
set_target_properties( dummy_withdep_app PROPERTIES LANGUAGE %s )
set_target_properties( dummy_withdep_app PROPERTIES EXPORT_COMPILE_COMMANDS ON )
target_link_libraries( dummy_withdep_app %s )

#foreach( tmp ${FAKE_USAGE_VAR_LIST} )
#  if ( DEFINED "${tmp}" )
#    set( "${tmp}" "${${tmp}}" )
#  endif()
#endforeach()

"""%( fpargs,args.language,args.language,args.deptargets )

    return cmake

def launch_cmake( args, blddir, srcdir ):
    cmd = get_cmake_command( args ) + [ srcdir ]
    cmd_str = ' '.join(shlex.quote(str(c)) for c in cmd)
    if args.verbose:
        print(f"Launching: {cmd_str}")
    def flush():
        ( sys.stdout.flush(), sys.stderr.flush() )
    flush()
    try:
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        out = p.communicate()[0]
        ec = p.returncode
    except ( OSError, subprocess.SubprocessError ):
        raise SystemExit('Error: CMake command failed')
    flush()
    if out and args.verbose:
        sys.stdout.buffer.write(out)
    if ec:
        raise SystemExit('Error: CMake command failed')


@contextlib.contextmanager
def changedir(subdir):
    """Context manager for working in a given subdir and then switching back"""
    the_cwd = os.getcwd()
    os.chdir(subdir)
    try:
        yield
    finally:
        os.chdir(the_cwd)

@contextlib.contextmanager
def work_in_tmpdir():
    """Context manager for working in a temporary directory (automatically
    created+cleaned) and then switching back"""
    import tempfile
    the_cwd = os.getcwd()
    try:
        with tempfile.TemporaryDirectory() as tmpdir:
            os.chdir(tmpdir)
            yield
    finally:
        os.chdir(the_cwd)

def actual_main( args ):
    blddir = ( pathlib.Path('.') / 'bld' ).absolute().resolve()
    srcdir = ( pathlib.Path('.') / 'src' ).absolute().resolve()
    blddir.mkdir()
    srcdir.mkdir()
    dummy_c_or_cpp_main = 'int main(void) { return 0; }\n'
    ( srcdir /'dummy_nodep.input' ).write_text( dummy_c_or_cpp_main )
    if args.findpackage:
        ( srcdir /'dummy_withdep.input' ).write_text( dummy_c_or_cpp_main )

    cmakeliststxt = create_cmakeliststxt( args )
    if args.verbose:
        print('Created CMakeLists.txt >>>\n\n')
        print(cmakeliststxt)
        print('\n\n<<< End of CMakeLists.txt')
    ( srcdir/'CMakeLists.txt' ).write_text(cmakeliststxt)


    with changedir(blddir):
        launch_cmake( args, blddir, srcdir )

    compile_json = json.loads( ( blddir / 'compile_commands.json').read_text() )
    assert len(compile_json)== (2 if args.findpackage else 1)

    def parse_links_txt( name ):
        return ParseCmd( (blddir / 'CMakeFiles'/f'dummy_{name}_app.dir'/'link.txt'
                          ).read_text().strip(),f'dummy_{name}.input' )
    def parse_compile_cmd( name ):
        ll = [ e for e in compile_json
               if ( pathlib.Path(e['file']).name
                    == f'dummy_{name}.input')][0]['command']
        return ParseCmd(ll,f'dummy_{name}.input')

    compile_nodep = parse_compile_cmd('nodep')
    link_nodep = parse_links_txt('nodep')

    out = dict( base = dict( compile
                             = dict( progname = compile_nodep.progname,
                                     args_preoutput = compile_nodep.args_preoutput,
                                     args_postoutput = compile_nodep.args_postoutput ),
                             link
                             = dict( progname = link_nodep.progname,
                                     args_preoutput = link_nodep.args_preoutput,
                                     args_postoutput = link_nodep.args_postoutput ) ) )
    if args.findpackage:
        compile_withdep = parse_compile_cmd('withdep')
        link_withdep = parse_links_txt('withdep')
        out['extra'] = dict( compileflags = compile_withdep.extra_args(compile_nodep),
                             linkflags = link_withdep.extra_args(link_nodep) )

    if args.output=='stdout':
        import pprint
        pprint.pprint(out)
        return
    out_json = json.dumps(out)

    if args.output=='stdout_json':
        print(out_json)
    else:
        args.output.write_text( out_json )

def main( argv = None ):
    args = parse_args( argv = argv )
    with work_in_tmpdir():
        actual_main( args )

if __name__ == '__main__':
    main()



from . import envcfg as _envcfg
import sys as _sys

black = '\033[30m'
red = '\033[31m'
green = '\033[32m'
yellow = '\033[33m'
blue = '\033[34m'
magenta = '\033[35m'
cyan = '\033[36m'
lightgrey = '\033[37m'
darkgrey = '\033[90m'
lightred = '\033[91m'
lightgreen = '\033[92m'
lightyellow = '\033[93m'
lightblue = '\033[94m'
lightmagenta = '\033[95m'#pink
lightcyan = '\033[96m'
white = '\033[97m'
end = '\033[0m'

ok = lightgreen
bad = lightred
bldmsg_pkg = lightblue
bldmsg_objectfile = white
bldmsg_symlink = lightyellow
bldmsg_headers = lightyellow
bldmsg_shlib = lightmagenta
bldmsg_app = lightgreen
bldmsg_pymod = lightcyan
bldmsg_global = lightblue
bldmsg_notallselectwarn = lightyellow
grep_match = lightblue
grep_unmatch = magenta
inc_samepkg = blue
inc_samedir = lightblue
#inc_dynpkg = lightcyan
warnenvsetup = lightyellow

#allow users to modify colours by having stuff like "export SIMPLEBUILD_COLOR_FIX=bldmsg_symlink=red" in their environment:
cf = _envcfg.var.color_fix_code
if cf:
    exec(cf)

#Disable colours when output is redirected:
if not _sys.stdout.isatty():
    for cname in [c for c in dir() if not c[0]=='_']:
        exec('%s=""'%cname)


if __name__=='__main__':
    for cname in [c for c in dir() if not c[0]=='_' and not c=='end']:
        col=eval(cname)
        print(''.join([cname.ljust(20),' : ',str(col),cname,end]))

def bldcol(name):
    name='bldmsg_'+name
    assert name in globals()
    return '${COL_%s}'%name.upper()
bldend='${COL_END}'



from . import target_base
from . import col
import os
import re
from . import dirs
from . import utils
from . import error
from . import conf

_patterns=[]
#dirtypes=set()
dirtype_info = {}
def register_pattern(pattern,target_factory):
    global _patterns,dirtypes_info#,dirtypes
    _patterns+=[pattern]
    dt=pattern[0:4]
    dirtype_info[dt] = (re.compile('^'+pattern+'$').match,target_factory)
    assert len(dirtype_info)==len(_patterns),"patterns not uniquely defined by first 4 chars"

for p,f in conf.target_factories_for_patterns():
    register_pattern(p,f)
nonpattern_tfacts=conf.target_factories()

def apply_patterns(pkg):
    targets=[]
    dn=dirs.pkg_dir(pkg)
    dirtypes=set()
    for d in os.listdir(dn):
        dirtype=d[0:4]
        nfo = dirtype_info.get(dirtype,None)
        dd=os.path.join(dn,d)
        if nfo and nfo[0](d) and os.path.isdir(dd):
            thefactory = nfo[1]
            ll=thefactory(pkg,d)#call factory
            #subdirs matching a target-generating pattern can not have subdirs themselves. This seems like a good place to check for it:
            for sf in os.listdir(dd):
                if sf=='.ipynb_checkpoints':
                    continue#ignore these as a workaround (although we should guard against them being committed)
                sf=os.path.join(dd,sf)
                if os.path.isdir(sf):
                    if os.path.basename(sf).startswith('__pycache__'):
                        continue
                    error.error('Sub-directories are not allowed in %s/%s/'%(pkg.name,d))
            if ll:
                dirtypes.add(dirtype)
                targets += ll
    return targets,dirtypes

def create_pkg_targets(pkg):
    assert not hasattr(pkg,'targets')
    #create targets for package
    targets=[]

    #find subdirs matching patterns and apply their associated target factory:
    pt,dt=apply_patterns(pkg)
    targets += pt

    for tf in nonpattern_tfacts:
        targets+=tf(pkg,dt)

    #Packages always have special done target which depends on all the other targets:
    targets += [target_base.TargetPkgDone(pkg,targets)]

    pkg.targets = targets

    d=dirs.pkg_cache_dir(pkg)
    utils.mkdir_p(d)
    utils.pkl_dump(targets,os.path.join(d,'targets.pkl'))

def create_pkg_targets_from_pickle(pkg):
    assert not hasattr(pkg,'targets')
    pkg.targets = utils.pkl_load(dirs.pkg_cache_dir(pkg,'targets.pkl'))

class TargetGlobalSysModules(target_base.Target):
    isglobal=True
    contains_message=True
    def __init__(self):
        self.name='global__sysmods'
        self.pkgname = None
        self.deps = []
        self.code = [
            '@if [ ${VERBOSE} -ge 0 ]; then echo "%sInstalling global system modules%s"; fi'%(col.bldcol('global'),col.bldend),
            'mkdir -p ${INST}/python/simplebuild',
            'touch ${INST}/python/simplebuild/__init__.py'
        ]
        self.deps += [
            '${BLD}/cfg.py',
            '${BLD}/_cfg_data.pkl'
        ]
        self.code += [
            'cp -f ${BLD}/cfg.py ${INST}/python/simplebuild/cfg.py',
            'cp -f ${BLD}/_cfg_data.pkl ${INST}/python/simplebuild/_cfg_data.pkl'
        ]


def create_global_targets():
    return [
        TargetGlobalSysModules(),
    ]



#Standalone utilities, which can be safely used without side-effects (e.g. for
#sphinx extensions):
import re

def _find_pkg_dirs_under_basedir( basedir,
                                  cfgfilename,
                                  error_fct = None ):
    def _error( msg ):
        if error_fct is None:
            from . import error
            error.error(msg)
        else:
            error_fct(msg)

    if not basedir.is_dir():
        _error('Asked to search non-existing directory for'
               f' packages: {basedir}')

    #Ignore simplebuild's own cache dirs:
    if any( ( basedir / fn ).exists()
            for fn in ('.sbinstalldir','.sbbuilddir') ):
        return []
    def _norm( p ):
        return p.absolute().resolve()
    if ( basedir / cfgfilename ).exists():
        #basedir is itself a pkg dir
        return [ _norm(basedir) ]

    #dir is not a pkg dir so check sub-dirs:
    pkg_dirs = []
    seen_lc = set()
    for p in basedir.iterdir():
        pl = p.name.lower()
        if not pl or pl[0]=='.' or '~' in pl or '#' in pl:
            continue#always ignore hidden dirs and weird files
        if pl in seen_lc:
            _error('Directory (and file) names differing only in'
                   ' casing are not allowed, due to being a potential'
                   ' source of error on some file systems. \nProblem'
                   f' occured with {p.name} in the directory {basedir}')
        seen_lc.add( pl )
        if p.is_dir():
            pkg_dirs+=_find_pkg_dirs_under_basedir(p,
                                                   cfgfilename,
                                                   error_fct)
    return pkg_dirs

def find_pkg_dirs_under_basedir( basedir, error_fct = None ):
    if error_fct is None:
        def error_fct( msg ):
            raise RuntimeError(msg)
    return _find_pkg_dirs_under_basedir( basedir, 'pkg.info', error_fct )

_nameval_pattern='^[a-zA-Z][a-zA-Z0-9_]{2,24}$'
_nameval=re.compile(_nameval_pattern).match
_nameval_descr=( f'Valid names follow pattern {_nameval_pattern}, contain no'
                 ' duplicate underscores and is not "_simple_build_system"' )

def pkgname_valid(n):
    return _nameval(n) and '__' not in n and n != '_simple_build_system'

def pkgname_valid_or_errmsg(n):
    if not pkgname_valid(n):
        return 'Invalid package name: "%s". %s.'%(n,_nameval_descr)



from . import target_base
from . import dirs
from . import utils

class TargetLibAvail(target_base.Target):
    def __init__(self,pkg):
        has_libsrc = pkg.has_lib()
        pklfile=dirs.pkg_cache_dir(pkg,'libavail.pkl')
        pklcont=has_libsrc
        mpf=dirs.makefile_blddir(pklfile)
        utils.update_pkl_if_changed(pklcont,pklfile)
        self.name='%s_libavail'%pkg.name
        self.pkgname=pkg.name
        self.deps=[mpf]
        self.code = []
        if has_libsrc:
            self.deps+=['%s_libsrc'%pkg.name]

def tfactory_libavail(pkg,dirtypes):
    #a silly little target whose only purpose in life is to make sure that
    #clients are relinked when a library disappears.
    #
    #(alternatively we could create an ldflag file and depend on it in bin targets)
    return [TargetLibAvail(pkg)]



def perform_cfg( *,
                 select_filter=None,
                 force_reconf=False,
                 load_all_pkgs=False,
                 quiet=False,
                 verbose=False,
                 export_jsoncmds=False
                ):
    import os
    import sys
    import glob
    from . import extractenv
    from . import utils
    from . import loadpkgs
    from . import dirs
    from . import conf
    from . import error
    from . import mtime
    from . import envcfg
    import pipes
    import shutil
    from .io import print

    #Pre-inspect package directories, simply finding the package dirs for now (do it already to detect some errors early):
    all_pkgdirs = loadpkgs.find_pkg_dirs ( dirs.pkgsearchpath )

    #Inspect package tree and load the necessary pkg.info files, given the filters:
    pl = loadpkgs.PackageLoader( all_pkgdirs,
                                 select_filter,
                                 autodeps=conf.autodeps,
                                 load_all = load_all_pkgs )

    #Actual extdeps that might be used:
    actually_needed_extdeps = set()
    for p in pl.enabled_pkgs_iter():
        if p.enabled:
            actually_needed_extdeps.update(p.direct_deps_extnames)

    volatile_misc = envcfg.var.reconf_env_vars[:]

    #Inspect environment via cmake (caching the result of previous runs).

    def calc_current_reconf_environment( envdict ):
        return dict(
            install_dir = str(conf.install_dir()),
            build_dir = str(conf.build_dir()),
            command_paths = dict( (b,str(shutil.which(b)))
                                  for b in sorted(set(envdict['autoreconf']['bin_list'].split(';')))),
            environment_variables = dict( (e,os.environ.get(e))
                                          for e in sorted(set([*envdict['autoreconf']['env_list'].split(';'),*volatile_misc])) )
        )

    assert dirs.blddir.is_dir()
    assert ( dirs.blddir / '.sbbuilddir' ).exists()

    #TODO: This need to cache and retrigger on cmakeargs is a leftover from
    #ancient times. Should we simply remove it?
    cmakeargs = []

    envdict = None
    if not force_reconf and os.path.exists(dirs.envcache):
        #load from cache:
        envdict=utils.pkl_load(dirs.envcache)
        current_reconf_environment = calc_current_reconf_environment(envdict)
        #check if reconf has been triggered:
        if envdict['_cmakeargs'] != cmakeargs:
            if not quiet:
                print("Change in cmake variables detected => reinspecting environment")
            envdict=None
        elif envdict['_autoreconf_environment'] != current_reconf_environment:
            if not quiet:
                print("Change in environment detected => reinspecting via CMake")
                for e in calc_env_changes( envdict['_autoreconf_environment'],
                                           current_reconf_environment ):
                    print(f'  -> change: {e}')
            envdict=None
        elif len(actually_needed_extdeps - envdict['_actually_needed_extdeps']) > 0:
            if not quiet:
                print("Change in relevant extdeps detected => reinspecting via CMake")
            envdict=None

    #Make sure any change in reported python version triggers reconf (especially important during python2->python3 transision):
    pyversionstr = str(sys.version_info)
    if envdict and envdict.get('_pyversion',None)!=pyversionstr:
        if not quiet:
            print("Change in python environment detected => reinspecting via CMake")
            envdict=None

    from . import env
    _devel_testextract = os.environ.get('SBLD_DEVEL_TESTCMAKECFG_MODE')
    if _devel_testextract is not None:
        assert _devel_testextract in ('REUSE','INVOKECMAKE')

    if envdict:
        #Can reuse config:
        env.env=envdict
        if _devel_testextract is not None:
            assert _devel_testextract == 'REUSE'
    else:
        #must extract the hard way via cmake:
        if _devel_testextract is not None:
            assert _devel_testextract == 'INVOKECMAKE'

        envdict = extractenv.extractenv(
            cmakeargs = cmakeargs,
            actually_needed_extdeps = actually_needed_extdeps,
            quiet = quiet,
            verbose = verbose
        )
        if not envdict:
            utils.rm_f(dirs.envcache)
            error.error('Failure during cmake configuration')
        assert '_cmakeargs' not in envdict
        envdict['_cmakeargs']=cmakeargs
        assert '_actually_needed_extdeps' not in envdict
        envdict['_actually_needed_extdeps'] = actually_needed_extdeps
        assert '_autoreconf_environment' not in envdict
        envdict['_autoreconf_environment'] = calc_current_reconf_environment(envdict)
        envdict['_pyversion'] = pyversionstr
        envdict['system']['volatile'].update({ 'misc': {**dict((e,os.getenv(e)) for e in volatile_misc)} })
        utils.pkl_dump(envdict,dirs.envcache)
        #configuration fully extracted:
        env.env=envdict

        #(re)create any libs_to_symlink whenever extracting env via cmake:
        dirs.create_install_dir()
        linkdir = dirs.installdir / 'lib' / 'links'
        assert ( dirs.installdir / '.sbinstalldir' ).exists()
        shutil.rmtree( linkdir, ignore_errors=True)
        linkdir.mkdir(parents=True)
        for lib_to_symlink in env.env['system']['runtime']['libs_to_symlink']:
            #Symlink not only the given file, but also all other files in target
            #dir with same basename and pointing to same file (e.g. if libA.so
            #and libA.x.so both point to libA.so.x.y, then add symlinks for all
            #three names.
            lib_to_symlink=os.path.abspath(lib_to_symlink)
            bn,dn=os.path.basename(lib_to_symlink),os.path.dirname(lib_to_symlink)
            assert bn
            rp=os.path.realpath(lib_to_symlink)
            for f in glob.glob(os.path.join(dn,'%s*'%bn.split('.')[0])):
                if os.path.realpath(f)==rp:
                    os.symlink(rp,os.path.join(linkdir,os.path.basename(f)))

    #update extdep, language info and runtime files, to trigger makefile target rebuilds as necessary:
    #utils.mkdir_p(os.path.join(dirs.blddir,'extdeps'))
    utils.mkdir_p(dirs.blddir / 'extdeps')
    for extdep,info in envdict['extdeps'].items():
        utils.update_pkl_if_changed(info,dirs.blddir / 'extdeps' / extdep)

    utils.mkdir_p(dirs.blddir / 'langs')
    pybind11info = dict((k,v) for k,v in env.env['system']['general'].items() if k.startswith('pybind11'))
    for lang,info in envdict['system']['langs'].items():
        #add pybind11 info as well, to trigger rebuilds when needed.
        utils.update_pkl_if_changed( {'langinfo':info,'pybind11info':pybind11info},
                                     dirs.blddir / 'langs' / lang )

    #Possible external dependencies (based solely on available ExtDep_xxx.cmake files):
    possible_extdeps = set(envdict['extdeps'].keys())

    pl.check_no_forbidden_extdeps( possible_extdeps )


    #Gracefully disable packages based on missing external dependencies

    missing_extdeps = set(k for k,v in envdict['extdeps'].items() if not v['present'])
    missing_needed = set(missing_extdeps).intersection(pl.active_deps_extnames())

    if missing_needed:
        autodisable = True
        if autodisable:
            disabled = pl.disable_pkgs_with_extdeps(missing_extdeps)
            if disabled and not quiet:
                print(f'Disabled {len(disabled)} packages due '
                      'to missing external dependencies')
        else:
            #This is a leftover from when user absolutely wanted a particular
            #list of packages to be enabled. Perhaps we want to introduce such a
            #feature again?
            error.error('Selected packages have missing external '
                        'dependencies: "%s"'%'", "'.join(missing_needed))

    #Load old contents of global db:
    from . import db

    #Check if volatile parts of environment changed:
    volatile = env.env['system']['volatile']
    if db.db.get('volatile',None) != volatile:
        db.db['volatile']=volatile
        db.db['pkg2timestamp']={}#make all pkgs "dirty"

    #For each enabled package, put it in one of the following categories:
    #   a) Timestamps of package and all parents are unchanged
    #        => Do nothing for package (all info needed by other packages will be in the global
    #                                   db and the makefile for the package is still there).
    #   b) Timestamp of package unchanged but some parent had new timestamp.
    #        => Load pkg targets from pickle file (they still need to have setup() called).
    #   c) Timestamp of package itself changed.
    #        => Create targets for package from scratch.

    ts=db.db['pkg2timestamp']
    rd=db.db['pkg2reldir']

    #Force reconfig of pkg+clients if any direct extdep's changed!
    pkgdb_directextdeps = db.db['pkg2directextdeps']
    for p in pl.enabled_pkgs_iter():
        _e = pkgdb_directextdeps.get(p.name)
        if (_e or set()) != p.direct_deps_extnames:
            if verbose and _e is not None:
                print(f'Detected direct USEEXT changes in {p.name}')
            p.set_files_in_pkg_changed()
            #Not 100% sure we should do this here already, but it most likely
            #does not hurt:
            pkgdb_directextdeps[p.name] = set(e for e in p.direct_deps_extnames)

    #TODO: Implement the following for increased safety: We should first check
    #all packages if they moved reldirname. Those packages all have to be nuked
    #for safety, and if any packages are nuked, all must be reconfigured
    #(updating a symlink is not enough, as it could be that the compiler reads
    #the ultimate located of links and embeds the realpath in
    #executables). Moving packages should be rather rare, so best to be sure.

    for p in pl.enabled_pkgs_iter():
        mt=mtime.mtime_pkg(p)
        if not ( ts.get(p.name,None) == mt ):
            if verbose:
                print(f'Detected timestamp changes in {p.name}')
            p.set_files_in_pkg_changed()
            ts[p.name]=mt
        if rd.get(p.name,None)!=p.reldirname:
            if verbose:
                print(f'Detected location changes in {p.name}')
            p.set_files_in_pkg_changed()#added adhoc for safety
            #Make sure pkg symlink is in place (refering to the package in
            #symlinked location, potentially allows users to move pkgs around).
            lt=dirs.pkg_dir(p)
            if lt.is_dir() and not lt.is_symlink():
                assert utils.path_is_relative_to( lt, conf.build_dir() ) and ( conf.build_dir()/'.sbbuilddir' ).exists()
                shutil.rmtree( lt, ignore_errors=True)
            else:
                lt.parent.mkdir( parents = True, exist_ok = True )
                if lt.is_symlink():
                    lt.unlink()
                lt.symlink_to(p.dirname,target_is_directory=True)
            rd[p.name]=p.reldirname#ok to update immediately?

    def nuke_pkg(pkgname):
        if verbose:
            print(f'Completely wiping cache for {pkgname}')
        conf.uninstall_package(pkgname)
        utils.rm_rf(dirs.pkg_cache_dir(pkgname))
        utils.rm_rf(dirs.pkg_dir(pkgname))#remove link to pkg
        nt=pipes.quote(os.path.join(dirs.blddir,'named_targets',pkgname))
        utils.system('rm -f %s %s_*'%(nt,nt))#Fixme: this is potentially
                                             #clashy... we should not use single
                                             #underscores if they are allowed in
                                             #package names (see also fixmes in
                                             #conf.py).
        db.clear_pkg(pkgname)

    from . import target_builder
    any_enabled=False
    for p in pl.enabled_pkgs_iter():
        any_enabled=True
        if p.files_in_pkg_changed():
            #recreate pkg targets from scratch
            old_parts = db.db['pkg2parts'].get(p.name,set())
            db.clear_pkg(p.name)
            db.db['pkg2parts'][p.name]=set()
            err_txt,unclean_exception = None,None
            try:
                if verbose:
                    print(f'Recreating pkg targets for {p.name}')
                target_builder.create_pkg_targets(p)#this also dumps to pickle!
            except KeyboardInterrupt:
                err_txt = "Halted by user interrupt (CTRL-C)"
            except Exception as e:
                err_txt=str(e)
                if not err_txt:
                    err_txt='<unknown error>'
                if not isinstance(e,error.Error):
                    unclean_exception = e
            if err_txt:
                #Error! Nuke the offending package and mark ALL packages for
                #reconfig (since even the OK ones will not get makefiles updated).
                from . import col
                print()
                print(f'{col.bad}ERROR during configuration of package {p.name}:{col.end}')
                print()
                for e in err_txt.splitlines():
                    print(f'  {e}')
                print()
                nuke_pkg(p.name)
                db.db['pkg2timestamp']={}
                db.save_to_file()
                if unclean_exception:
                    #unknown exceptions need an ugly traceback so we can notice and try to avoid them in the future
                    error.print_traceback(unclean_exception)
                error.clean_exit(1)
            new_parts = db.db['pkg2parts'][p.name]
            disappeared_parts = old_parts.difference(new_parts)
            if disappeared_parts:
                conf.deinstall_parts(dirs.installdir,p.name,new_parts,disappeared_parts)
            #If we get here, package was updated OK:
            ts[p.name] = mtime.mtime_pkg(p)
            rd[p.name] = p.reldirname
            pkgdb_directextdeps[p.name] = set(e for e in p.direct_deps_extnames)
        elif p.any_parent_changed():
            #merely load targets from pickle!
            target_builder.create_pkg_targets_from_pickle(p)
        else:
            #nothing to be done
            pass

    if not any_enabled:
        error.error('No packages were enabled in the present configuration')

    #Global targets:
    global_targets = target_builder.create_global_targets()

    #Perform setup calls to targets:
    for p in pl.enabled_pkgs_iter():
        if hasattr(p,'targets'):
            for t in p.targets:
                assert not t.isglobal
                t.setup(pl.name2pkg[t.pkgname])

    for t in global_targets:
        assert t.isglobal
        t.setup(None)#has no associated pkg

    #Validate and collect:
    all_targets=[]
    for p in pl.enabled_pkgs_iter():
        if hasattr(p,'targets'):
            for t in p.targets:
                t.validate()
            all_targets += p.targets

    for t in global_targets:
        t.validate()
    all_targets+=global_targets

    export_jsoncmds_pkg = None
    export_jsoncmds_finalise = None
    if export_jsoncmds:
        from . import _export_jsoncmds
        export_jsoncmds_pkg      = _export_jsoncmds.export_pkg
        export_jsoncmds_finalise = _export_jsoncmds.finalise

    #setup makefiles:
    utils.mkdir_p(dirs.makefiledir)
    from . import makefile

    #the main makefile:
    enabled_pkgnames=list(p.name for p in pl.enabled_pkgs_iter())
    makefile.write_main(global_targets,enabled_pkgnames)


    #package makefiles:
    if export_jsoncmds_pkg:
        _makefile_var_map = makefile.get_makefile_global_vars()
    for p in pl.enabled_pkgs_iter():
        if hasattr(p,'targets'):#no targets means reuse old makefile
            makefile.write_pkg(p)
            if export_jsoncmds_pkg:
                export_jsoncmds_pkg(p,_makefile_var_map)

    if export_jsoncmds_finalise:
        export_jsoncmds_finalise(dirs.blddir / 'compile_commands.json')

    enabled_pkgnames=set(enabled_pkgnames)
    if 'enabled_pkgnames' in db.db:
        #Test if any pkgs got disabled compared to last time and remove it from install area:
        for disappeared_pkgname in db.db['enabled_pkgnames'].difference(enabled_pkgnames):
            if not quiet:
                print("Uninstalling package %s"%disappeared_pkgname)
            nuke_pkg(disappeared_pkgname)
    db.db['enabled_pkgnames'] = enabled_pkgnames

    #Save new state:
    db.save_to_file()


    dir_names = ('projdir','pkgsearchpath','blddir',
                 'installdir','testdir','envcache')
    dirdict = dict(((d, getattr(dirs, d)) for d in dir_names))

    #Update dynamic python module with information, if needed:
    cfgfile=dirs.blddir / 'cfg.py'
    cfg_picklefile=dirs.blddir / '_cfg_data.pkl'
    simplebuild_cfg_pkgs_dict = dict((p.name,p.info_as_dict()) for p in pl.pkgs)
    cfg_pickle_data = dict( cmakeargs = cmakeargs,
                            pkgs = simplebuild_cfg_pkgs_dict )
    utils.update_pkl_if_changed( cfg_pickle_data,
                                 cfg_picklefile )
    content='"""File automatically generated by simplebuild"""'
    content+=f"""
import pickle as _pkl
import pathlib as _pl

_pklfile = _pl.Path(__file__).parent / '{cfg_picklefile.name}'
with _pklfile.open('rb') as fh:
    _pkldata = _pkl.load(fh)
cmakeargs = _pkldata['cmakeargs']
pkgs = _pkldata['pkgs']
del _pklfile
del _pkldata

class Dirs:
"""
    for k,v in sorted(dirdict.items()):
      val = (f"_pl.Path({repr(str(v))}).absolute()" if not isinstance(v,list)
             else '['+','.join([f"_pl.Path({repr(str(el))})" for el in v])+']')
      content+=f"""
  @property
  def {k}(self):
    return {val}
"""
    content+="""
  @property
  def instdir(self):
    return _pl.Path(__file__).parent.parent.parent.absolute()

  @property
  def testrefdir(self):
    return self.instdir / 'tests' / 'testref'

  @property
  def datadir(self):
    return self.instdir / 'data'

  @property
  def libdir(self):
    return self.instdir / 'lib'

  @property
  def includedir(self):
    return self.instdir / 'include'

  @property
  def installprefix(self):
    return self.instdir

dirs = Dirs()
if __name__=="__main__":
  import pprint as pp
  pp.pprint(pkgs)
  pp.pprint({ p:getattr(dirs,p) for p in dir(dirs) if not p.startswith('_')})
"""
    try:
        current_content = cfgfile.read_text()
    except FileNotFoundError:
        current_content = None
    if current_content != content:
        cfgfile.write_text( content )

    return pl

def calc_env_changes( old, new ):
    if set(old.keys()) != set(new.keys()):
        yield 'autoreconf format change!'
        return
    dns = ('install_dir','build_dir')

    for dn in dns:
        if old[dn] != new[dn]:
            yield f'Directory changed: {dn}'
    for tn in [ tn for tn in sorted(old.keys()) if tn not in dns ]:
        old_tn, new_tn = old[tn], new[tn]
        ko, kn = set(old_tn.keys()), set(new_tn.keys())
        for k in kn|ko:
            vo, vn = old_tn[k], new_tn[k]
            if vo!=vn:
                vofmt = f'"{vo}"' if vo is not None else '<unset>'
                vnfmt = f'"{vn}"' if vn is not None else '<unset>'
                yield f'value of "{k}" ({tn}) changed from {vofmt} to {vnfmt}'
        for k in ko-kn:
            yield f'{tn} - no longer tracking "{k}"'
        for k in kn-ko:
            yield f'{tn} - now also tracking "{k}"'



from .singlecfg import SingleCfg
from . import error
import pathlib

def load_builtin_cfgs():
    pd = pathlib.Path(__file__).absolute().parent / 'data'
    cfgs = [ ( pd / 'pkgs-core' / 'simplebuild.cfg' ).absolute().resolve(),
             ( pd / 'pkgs-core_val' / 'simplebuild.cfg' ).absolute().resolve() ]
    return [ ( p, SingleCfg.create_from_toml_file(p,ignore_build=True) ) for p in cfgs ]

class CfgBuilder:

    """Class which, based on an initial "master" simplebuild.cfg, can locate and
    extract other simplebuild cfg's, and provide a combined configuration object.

    The "build" section (build mode, ncpu, pkgfilter, etc.) is taken straight
    from the master config, while the other sections are based on a combination
    of data in the master_cfg and any cfg's of the dependencies.

    """

    def __init__(self, master_cfg : SingleCfg, master_cfg_file ):
        #Input:
        #Take build settings straight from master_cfg:
        from . import io as _io
        print_prefix = f'{_io.print_prefix}cfgbuilder:'
        self.__print = lambda *a,**kw: _io.print_no_prefix(print_prefix,*a,**kw)
        self.__is_verbose = _io.is_verbose
        self.__build_mode = master_cfg.build_mode
        #self.__build_njobs = master_cfg.build_njobs
        self.__build_cachedir = master_cfg.build_cachedir
        self.__build_pkg_filter = master_cfg.build_pkg_filter
        #self.__build_extdep_ignore = master_cfg.build_extdep_ignore

        #Build up everything else recursively, starting from the master_cfg:
        self.__pkg_path = []#result 1
        self.__env_paths = {}#result 2
        self.__used_cfg_files = set()#make sure we don't consider the same cfg-file twice
        self.__used_cfg_files.add( master_cfg_file )
        self.__available_unused_cfgs = []#only needed during build up
        self.__cfg_names_used = set()
        self.__cfg_names_missing = set([ master_cfg.project_name ])
        self.__use_cfg( master_cfg, is_top_level = True )

        for cfg_file, cfg in load_builtin_cfgs():
            #core/core-val bundles always available directly from core
            #installation, but we only add them if they were not added already
            #via a search_path entry (to facilitate development of core pkgs
            #directly from a git clone):
            if cfg.project_name in self.__cfg_names_missing:
                self.__print_verbose(f'Using built-in cfg: {cfg_file}')
                assert cfg_file not in self.__used_cfg_files
                self.__used_cfg_files.add( cfg_file )
                self.__use_cfg( cfg )

        if self.__cfg_names_missing:
            #Only consider python plugins if there is anything we did not find
            #already in the directly requested search paths (that way a dev
            #environment with everything checked out will be unaffected by what
            #plugins happen to be installed in the environment):
            self.__consider_python_plugins()
        if self.__cfg_names_missing:
            _p='", "'.join(self.__cfg_names_missing)
            _s = 's' if len(self.__cfg_names_missing)>2 else ''
            error.error('Could not find dependent project%s: "%s"'%(_s,_p))
        self.__pkg_path = tuple( self.__pkg_path )
        del self.__available_unused_cfgs

    def __print_verbose( self, *a,**kw ):
        if self.__is_verbose():
            self.__print( *a, **kw )

    @property
    def build_mode( self ):
        return self.__build_mode

    #@property
    #def build_njobs( self ):
    #    return self.__build_njobs

    @property
    def build_cachedir( self ):
        return self.__build_cachedir

    @property
    def build_pkg_filter( self ):
        return self.__build_pkg_filter

    #@property
    #def build_extdep_ignore( self ):
    #    return self.__build_extdep_ignore

    @property
    def pkg_path(self):
        return self.__pkg_path

    @property
    def env_paths(self):
        return self.__env_paths

    def __use_cfg( self, cfg : SingleCfg, is_top_level = False ):
        assert cfg.project_name in self.__cfg_names_missing
        _cfgname='master-' if is_top_level else ''
        self.__print_verbose(f'Using {_cfgname}cfg from {cfg._cfg_file}')
        self.__cfg_names_missing.remove( cfg.project_name )
        self.__cfg_names_used.add( cfg.project_name )
        #Add dependencies and cfgs available in search paths:
        for cfgname in cfg.depend_projects:
            if cfgname not in self.__cfg_names_used:
                self.__cfg_names_missing.add( cfgname )
        for sp in cfg.depend_search_path:
            if not sp.exists():
                error.error
            assert sp.is_file()
            sp = sp.absolute()
            if sp in self.__used_cfg_files:
                continue
            self.__used_cfg_files.add( sp )
            depcfg = SingleCfg.create_from_toml_file(
                sp,
                ignore_build = not is_top_level
            )
            self.__available_unused_cfgs.append( depcfg )

        #Add actual pkg-dirs and env-path requests from cfg:
        if cfg.project_pkg_root not in self.__pkg_path:
            self.__pkg_path.append( cfg.project_pkg_root )
        for pathvar,contents in cfg.project_env_paths.items():
            if pathvar not in self.__env_paths:
                self.__env_paths[pathvar] = set()
            self.__env_paths[pathvar].update( contents )

        #Use cfgs we found as appropriate:
        self.__search_available_cfgs()

    def __search_available_cfgs( self ):
        if not self.__cfg_names_missing:
            return
        to_use_idx, to_use_cfg = set(), []
        for i,ucfg in enumerate(self.__available_unused_cfgs):
            if ucfg.project_name in self.__cfg_names_missing:
                to_use_cfg.append( ucfg )
                to_use_idx.add( i )
        if to_use_cfg:
            for i in sorted(to_use_idx,reverse=True):
                del self.__available_unused_cfgs[i]
            for ucfg in to_use_cfg:
                self.__use_cfg( ucfg )

    def __consider_python_plugins( self ):
        #First, find all possible python packages with plugins:
        own_pkg_name = pathlib.Path(__file__).parent.name
        import importlib
        import pkgutil
        possible_pyplugins = set(
            name
            for finder, name, ispkg
            in pkgutil.iter_modules()
            if ( name != own_pkg_name and
                 ( name.startswith('simplebuild_')
                   or name.startswith('_simplebuild_')
                  )
                )
        )

        #Load their cfgs:
        for name in sorted(possible_pyplugins):
            modname=f'{name}.simplebuild_bundle_list'
            self.__print_verbose(f'Trying python plugin module {modname}')
            try:
                mod = importlib.import_module(modname)
            except ModuleNotFoundError:
                self.__print_verbose(' -> skipping due to ModuleNotFoundError')
                continue
            if not hasattr( mod, 'simplebuild_bundle_list' ):
                self.__print_verbose(' -> skipping due to missing simplebuild_bundle_list function')
                continue
            srcdescr = 'Python module %s'%name
            for cfg_file in mod.simplebuild_bundle_list():
                if not cfg_file.is_absolute() or not cfg_file.is_file():
                    error.error(f'Non-absolute or non-existing cfg file path returned from {srcdescr}')
                #The .resolve() in the next line means that it is easier to work
                #with an on-disk clone of a repo which is otherwise providing
                #it's bundles after a pip install. If that clone is then "pip
                #install -e" (editable), then the resolve will go to the on-disk
                #clone rather than the symlinked file in the site-packages
                #area. The advantage is that newly added files in these repos
                #will still be picked up.
                cfg_file = cfg_file.absolute().resolve()
                if cfg_file in self.__used_cfg_files:
                    self.__print_verbose(f' -> skipping provided file already used: {cfg_file}')
                    continue#Already used
                self.__used_cfg_files.add( cfg_file )
                cfg = SingleCfg.create_from_toml_file( cfg_file, ignore_build = True )
                self.__available_unused_cfgs.append( cfg )

        #Use cfgs we found as appropriate:
        self.__search_available_cfgs()
